import csv
import pathlib
import shutil
import tempfile

import pandas as pd

from r2x.upgrader import upgrade_handler
from r2x.upgrader.functions import add_column, apply_header, melt, move_file, rename, set_index


def test_rename(tmp_path):
    test_file = tmp_path.joinpath("test.csv")
    test_file.touch()

    new_file = rename(test_file, "text.csv")
    assert isinstance(new_file, pathlib.Path)
    assert new_file.is_file()
    assert new_file.name == "text.csv"
    assert new_file.resolve().parent == test_file.resolve().parent


def test_move_file(tmp_path):
    test_file = tmp_path.joinpath("test.csv")
    test_file.touch()
    with tempfile.TemporaryDirectory() as tmpdirname:
        output_dir = pathlib.Path(tmpdirname)
        new_file = move_file(test_file, output_dir / "test.csv")
        assert new_file is not None
        assert new_file.is_file()
        assert new_file.name == "test.csv"
        assert new_file.resolve().parent == output_dir.resolve()


def test_melt(caplog):
    """Test the melt function."""
    temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    pd.DataFrame({"i": [1, 2], "r": [3, 4], "Q1": [10, 20], "Q2": [30, 40]}).to_csv(temp_file, index=False)

    melted_data = melt(temp_file)
    assert melted_data.equals(
        pd.DataFrame(
            {
                "i": [1, 2, 1, 2],
                "r": [3, 4, 3, 4],
                "quarter": ["Q1", "Q1", "Q2", "Q2"],
                "value": [10, 20, 30, 40],
            }
        )
    ), "Default melt operation failed"

    temp_file.unlink()

    # Test skip with already melted file
    temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    pd.DataFrame(
        {
            "i": [1, 2, 1, 2],
            "r": [3, 4, 3, 4],
            "quarter": ["Q1", "Q1", "Q2", "Q2"],
            "value": [10, 20, 30, 40],
        }
    ).to_csv(temp_file, index=False)
    melted_data = melt(temp_file)
    assert "has been already melted" in caplog.text


def test_apply_header():
    temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    with open(temp_file, mode="w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(["A", "B"])  # Header row
        writer.writerow([1, 3])  # Data row 1
        writer.writerow([2, 4])  # Data row 2

    updated_df = apply_header(temp_file, "x,y")
    assert updated_df is not None
    assert updated_df.equals(pd.DataFrame({"x": [1, 2], "y": [3, 4]})), "Header application failed"
    temp_file.unlink()

    temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    with open(temp_file, mode="w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(["A", "B"])  # Header row
        writer.writerow([1, 3])  # Data row 1
        writer.writerow([2, 4])  # Data row 2
    result = apply_header(temp_file, "a,b")
    assert result is None, "Header should not change but returned a DataFrame"

    updated_df_2 = apply_header(temp_file, "x,y")
    assert updated_df_2 is not None
    assert updated_df_2.equals(pd.DataFrame({"x": [1, 2], "y": [3, 4]})), "Second header application failed"

    temp_file.unlink()


def test_set_index():
    temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    with open(temp_file, mode="w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(["index", "A", "B"])  # Header row
        writer.writerow([0, 1, 3])  # Data row 1
        writer.writerow([1, 2, 4])  # Data row 2

    updated_df = set_index(temp_file, "test_index")
    assert updated_df is None
    temp_file.unlink()

    temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    with open(temp_file, mode="w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(["", "B"])  # Header row
        writer.writerow([1, 3])  # Data row 1
        writer.writerow([2, 4])  # Data row 2
    result = set_index(temp_file, "index")
    assert result is not None
    assert result.index.name == "index"
    temp_file.unlink()


def test_add_column():
    temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    with open(temp_file, mode="w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(["eall", "i", "v", "r", "t", "value"])  # Header row
        writer.writerow(["CO2", "beccs_mod", "new1", "p1", 2029, -0.918])
    result = add_column(temp_file, column_name="etype", default_value="combustion")
    assert result is not None
    assert "etype" in result.columns
    assert all(value == "combustion" for value in result["etype"])

    temp_file.unlink()
    temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    with open(temp_file, mode="w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(["eall", "etype", "i", "v", "r", "t", "value"])  # Header row
        writer.writerow(["CO2", "combustion", "beccs_mod", "new1", "p1", 2029, -0.918])
    result = add_column(temp_file, column_name="etype", default_value="combustion")
    assert result is None
    temp_file.unlink()


def test_upgrade_handler(tmp_path, reeds_data_folder):
    # Move data to new folder to not create backup on the main repo
    reeds_tmp_path = tmp_path / "reeds_folder"
    shutil.copytree(reeds_data_folder, tmp_path / "reeds_folder")

    upgrade_handler(reeds_tmp_path)
    pass



"""Plugin to add CCS incentive to the model.

This plugin is only applicable for ReEDs, but could work with similarly arrange data
"""

import polars as pl
from loguru import logger

from r2x.api import System
from r2x.units import ureg
from r2x.config_scenario import Scenario
from r2x.models.generators import Generator
from r2x.parser.handler import BaseParser


def update_system(
    config: Scenario,
    system: System,
    parser: BaseParser | None = None,
) -> System:
    """Apply CCS incentive to CCS eligible technologies.

    The incentive is calculated with the capture incentive ($/ton) and capture rate
    (ton/MWh), to produce a subtractor ($/MWh) implemented with PLEXOS' "Use of
    Service Charge".

    Parameters
    ----------
    config : Scenario
        The scenario configuration.
    parser : BaseParser
        The parser object used for parsing.
    system : System
        The system object to be updated.

    Notes
    -----
    The names of some of the columns for the parser data are specified in the `reeds_us_mapping.json`.
    """
    if not config.output_model == "plexos" and not config.input_model == "reeds-US":
        msg = "Plugin `ccs_credit.py` is not compatible with a model that is not Plexos or ReEDs input."
        raise NotImplementedError(msg)

    if parser is None:
        msg = "Missing parser information for ccs_credit. Skipping plugin."
        logger.debug(msg)
        return system

    required_files = ["co2_incentive", "emission_capture_rate", "upgrade_link"]
    if parser is not None:
        if not all(key in parser.data for key in required_files):
            logger.warning("Missing required files for ccs_credit. Skipping plugin.")
            return system

    production_rate = parser.data["emission_capture_rate"]

    # Some technologies on ReEDS are eligible for incentive but have not been upgraded yet. Since the
    # co2_incentive does not capture all the possible technologies, we get the technologies before upgrading
    # and if they exist in the system we apply the incentive.
    incentive = parser.data["co2_incentive"].join(
        parser.data["upgrade_link"], left_on="tech", right_on="to", how="left"
    )
    ccs_techs = incentive["tech"].unique()
    ccs_techs = ccs_techs.unique().extend(incentive["from"].unique())

    for generator in system.get_components(
        Generator, filter_func=lambda gen: gen.ext and gen.ext["reeds_tech"] in ccs_techs
    ):
        reeds_tech = generator.ext["reeds_tech"]
        reeds_vintage = generator.ext["reeds_vintage"]
        reeds_tech_mask = (
            (pl.col("tech") == reeds_tech)
            & (pl.col("region") == generator.bus.name)
            & (pl.col("vintage") == reeds_vintage)
        )
        generator_production_rate = production_rate.filter(reeds_tech_mask)

        if generator_production_rate.is_empty():
            msg = f"Generator {generator.name=} does not appear on the production rate file. Skipping it."
            logger.debug(msg)
            continue

        upgrade_mask = (
            (pl.col("from") == reeds_tech)
            & (pl.col("region") == generator.bus.name)
            & (pl.col("vintage") == reeds_vintage)
        )
        generator_incentive = incentive.filter(reeds_tech_mask.or_(upgrade_mask))["incentive"].item()
        generator.ext["UoS Charge"] = ureg.Quantity(
            -generator_incentive * generator_production_rate["capture_rate"].item(),
            "usd/MWh",  # Negative quantity to capture incentive in the objetive function.
        )
    return system



"""Augment results from CEM with PCM defaults."""

from argparse import ArgumentParser

from infrasys.base_quantity import BaseQuantity
import pandas as pd
from loguru import logger

from r2x.models import Generator
from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.parser.handler import BaseParser
from r2x.units import ActivePower
from r2x.utils import read_json


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--pcm-defaults-fpath",
        help="File containing the defaults",
    )


def update_system(
    config: Scenario, parser: BaseParser, system: System, pcm_defaults_fpath: str | None = None
) -> System:
    """Augment data model using PCM defaults dictionary.

    Parameters
    ----------
    config
        Scenario configuration class
    system
        InfraSys system
    pcm_defaults_path
        Path for json file containing the PCM defaults.

    Returns
    -------
        System
    """
    logger.info("Augmenting generators attributes")
    assert config.input_config
    if pcm_defaults_fpath is None:
        logger.debug("Using {}", config.input_config.defaults["pcm_defaults_fpath"])
        pcm_defaults = read_json(config.input_config.defaults["pcm_defaults_fpath"])
    else:
        logger.debug("Using custom defaults from {}", pcm_defaults_fpath)
        pcm_defaults: dict = read_json(pcm_defaults_fpath)

    reference_data = (
        pd.DataFrame.from_dict(pcm_defaults)
        .transpose()
        .reset_index()
        .rename(
            columns={
                "index": "tech",
            }
        )
    )
    reference_data.loc[reference_data.tech.str.startswith("battery"), "mean_time_to_repair"] = (
        config.input_config.defaults["storage_mean_time_to_repair"]
    )
    reference_data.loc[reference_data.tech.str.startswith("hyd"), "mean_time_to_repair"] = (
        config.input_config.defaults["hydro_mean_time_to_repair"]
    )
    reference_data.loc[reference_data.tech.str.startswith("hyd"), "max_ramp_up_percentage"] = (
        config.input_config.defaults["hydro_ramp_rate"] * 100
    )
    reference_data.loc[reference_data.tech.str.endswith("nd"), "min_stable_level_percentage"] = 1

    # Rename columns on the file itself and remove this
    reference_data = reference_data.loc[
        :,
        [
            "tech",
            "min_stable_level_percentage",
            "start_cost_per_MW",
            "max_ramp_up_percentage",
            "mean_time_to_repair",
            "min_down_time",
            "min_up_time",
            "maintenance_rate",
            "forced_outage_rate",
        ],
    ]
    dtypes = {
        "min_stable_level_percentage": "float",
        "start_cost_per_MW": "float",
        "max_ramp_up_percentage": "float",
        "mean_time_to_repair": "float",
        "min_down_time": "float",
        "min_up_time": "float",
        "maintenance_rate": "float",
        "forced_outage_rate": "float",
    }
    reference_data = reference_data.astype(dtypes)

    reference_techs = reference_data.tech.unique()

    for component in system.get_components(Generator):
        reeds_tech = getattr(component, "ext").get("reeds_tech")

        if reeds_tech not in reference_techs:
            continue

        wecc_data_row = (
            reference_data.loc[reference_data.tech.str.startswith(reeds_tech)]
            .dropna(axis=1)
            .to_dict(orient="records")[0]
        )
        values_to_add = {}

        # I do not like this implementation.
        values_to_add["planned_outage_rate"] = (
            getattr(component, "planned_outage_rate", None)
            or BaseQuantity(wecc_data_row.get("maintenance_rate"), "%")
            if wecc_data_row.get("maintenance_rate")
            else None
        )
        values_to_add["forced_outage_rate"] = (
            getattr(component, "forced_outage_rate", None)
            or BaseQuantity(wecc_data_row.get("forced_outage_rate"), "%")
            if wecc_data_row.get("forced_outage_rate")
            else None
        )
        values_to_add["ramp_up"] = (
            getattr(component, "active_power")
            * BaseQuantity(wecc_data_row.get("max_ramp_up_percentage"), "1/min")
            if wecc_data_row.get("max_ramp_up_percentage")
            else None
        )
        values_to_add["ramp_down"] = (
            getattr(component, "active_power")
            * BaseQuantity(wecc_data_row.get("max_ramp_up_percentage"), "1/min")
            if wecc_data_row.get("max_ramp_up_percentage")
            else None
        )
        values_to_add["min_rated_capacity"] = (
            getattr(component, "active_power")
            * BaseQuantity(wecc_data_row.get("min_stable_level_percentage"), "")
            if wecc_data_row.get("min_stable_level_percentage")
            else ActivePower(0, "MW")
        )
        values_to_add["mean_time_to_repair"] = (
            BaseQuantity(wecc_data_row.get("mean_time_to_repair"), "h")
            if wecc_data_row.get("mean_time_to_repair")
            else None
        )
        values_to_add["min_up_time"] = (
            BaseQuantity(wecc_data_row.get("min_up_time"), "h") if wecc_data_row.get("min_up_time") else None
        )
        values_to_add["min_down_time"] = (
            BaseQuantity(wecc_data_row.get("min_down_time"), "h")
            if wecc_data_row.get("min_down_time")
            else None
        )
        values_to_add["startup_cost"] = (
            getattr(component, "active_power")
            * BaseQuantity(wecc_data_row.get("start_cost_per_MW"), "usd/MW")
            if wecc_data_row.get("start_cost_per_MW")
            else None
        )
        valid_fields = {key: value for key, value in values_to_add.items() if key in component.model_fields}
        for key, value in valid_fields.items():
            setattr(component, key, value)
    return system



"""Useful check functions for upgrader."""

from pathlib import Path

import h5py
import pandas as pd
from loguru import logger


def check_if_h5_is_pandas_format(fpath: Path) -> bool:
    """Check if an h5 file is formatted using pandas."""
    try:
        _ = pd.read_hdf(fpath)
        return True
    except ValueError:
        return False


def check_if_h5_has_correct_index_names(fpath: Path) -> bool:
    """Check if has correct index names."""
    required_columns_1 = {"index_datetime", "index_year", "columns", "index_names", "data"}
    required_columns_2 = {"index_datetime", "columns", "index_names", "data"}
    with h5py.File(fpath, "r") as f:
        file_keys = set(f.keys())
    return required_columns_1.issubset(file_keys) or required_columns_2.issubset(file_keys)


def check_if_columm_is_datetime(fpath: Path, column: str = "index_datetime") -> bool:
    """Check if a h5 file has a datetime column."""
    with h5py.File(fpath, "r") as f:
        file_keys = set(f.keys())
        if column not in file_keys:
            logger.debug("`index_datetime` column not found on {}", fpath)
            return False
        first_row = f[column][:1]

    try:
        _ = pd.to_datetime(first_row.astype("U"))
    except:  # noqa: E722
        return False
    return True



import pandas as pd

from r2x.upgrader.checks import (
    check_if_columm_is_datetime,
    check_if_h5_has_correct_index_names,
    check_if_h5_is_pandas_format,
)
from r2x.upgrader.helpers import add_datetime_index, pandas_to_h5py, rename_index_names_from_h5


def test_pandas_checks(pandas_h5_weather_year):
    fpath = pandas_h5_weather_year
    assert check_if_h5_is_pandas_format(fpath)


def test_convert_weather_year_from_pandas(pandas_h5_weather_year, tmp_path):
    output_fpath = tmp_path / "output.h5"

    original_h5 = pd.read_hdf(pandas_h5_weather_year)

    assert pandas_to_h5py(original_h5, output_fpath)
    assert output_fpath.exists()
    assert check_if_columm_is_datetime(output_fpath)
    assert check_if_h5_has_correct_index_names(output_fpath)


def test_convert_solve_year_weather_year_from_pandas(pandas_h5_solve_year_and_weather_year, tmp_path):
    output_fpath = tmp_path / "output.h5"
    original_h5 = pd.read_hdf(pandas_h5_solve_year_and_weather_year)
    assert pandas_to_h5py(original_h5, output_fpath)
    assert output_fpath.exists()
    assert check_if_columm_is_datetime(output_fpath)
    assert check_if_h5_has_correct_index_names(output_fpath)


def test_convert_h5_no_datetime(h5_without_index_names, tmp_path):
    output_fpath = h5_without_index_names
    assert output_fpath.exists()
    assert not check_if_h5_has_correct_index_names(output_fpath)
    assert rename_index_names_from_h5(output_fpath)
    assert check_if_h5_has_correct_index_names(output_fpath)


def test_add_datetime_index(h5_with_index_names_no_datetime, tmp_path):
    output_fpath = h5_with_index_names_no_datetime
    assert output_fpath.exists()
    assert check_if_h5_has_correct_index_names(output_fpath)
    assert not check_if_columm_is_datetime(output_fpath)
    assert add_datetime_index(output_fpath)
    assert check_if_columm_is_datetime(output_fpath)



"""Models that capture topology types."""

from typing import Annotated

from pydantic import Field, NonNegativeFloat, PositiveInt

from r2x.enums import ACBusTypes
from r2x.models.core import BaseComponent
from r2x.models.named_tuples import MinMax
from r2x.units import Voltage, ureg


class Topology(BaseComponent):
    """Abstract type to represent the structure and interconnectedness of the system."""


class AggregationTopology(Topology):
    """Base class for area-type components."""


class Area(AggregationTopology):
    """Collection of buses in a given region."""

    peak_active_power: Annotated[NonNegativeFloat, Field(description="Peak active power in the area")] = 0.0
    peak_reactive_power: Annotated[NonNegativeFloat, Field(description="Peak reactive power in the area")] = (
        0.0
    )
    load_response: Annotated[
        float,
        Field(
            description=(
                "Load-frequency damping parameter modeling how much the load in the area changes "
                "due to changes in frequency (MW/Hz)."
            )
        ),
    ] = 0.0

    @classmethod
    def example(cls) -> "Area":
        return Area(name="New York")


class LoadZone(AggregationTopology):
    """Collection of buses for electricity price analysis."""

    peak_active_power: Annotated[NonNegativeFloat, Field(description="Peak active power in the area")] = 0.0
    peak_reactive_power: Annotated[NonNegativeFloat, Field(description="Peak reactive power in the area")] = (
        0.0
    )

    @classmethod
    def example(cls) -> "LoadZone":
        return LoadZone(name="ExampleLoadZone")


class Bus(Topology):
    """Abstract class for a bus."""

    number: Annotated[PositiveInt, Field(description="A unique bus identification number.")]
    bustype: Annotated[ACBusTypes, Field(description="Type of category of bus,")] | None = None
    area: Annotated[Area, Field(description="Area containing the bus.")] | None = None
    load_zone: Annotated[LoadZone, Field(description="the load zone containing the DC bus.")] | None = None
    voltage_limits: Annotated[MinMax, Field(description="the voltage limits")] | None = None
    base_voltage: (
        Annotated[Voltage, Field(gt=0, description="Base voltage in kV. Unit compatible with voltage.")]
        | None
    ) = None
    magnitude: (
        Annotated[NonNegativeFloat, Field(description="Voltage as a multiple of base_voltage.")] | None
    ) = None


class DCBus(Bus):
    """Power-system DC Bus."""

    @classmethod
    def example(cls) -> "DCBus":
        return DCBus(
            number=1,
            name="ExampleBus",
            load_zone=LoadZone.example(),
            area=Area.example(),
            base_voltage=20 * ureg.kV,
            bustype=ACBusTypes.PV,
        )


class ACBus(Bus):
    """Power-system AC bus."""

    angle: Annotated[float | None, Field(description="Angle of the bus in radians.", gt=-1.571, lt=1.571)] = (
        None
    )

    @classmethod
    def example(cls) -> "ACBus":
        return ACBus(
            number=1,
            name="ExampleBus",
            load_zone=LoadZone.example(),
            area=Area.example(),
            base_voltage=13 * ureg.kV,
            voltage_limits=MinMax(min=0.9, max=1.1),
            bustype=ACBusTypes.PV,
        )


class Arc(Topology):
    """Topological directed edge connecting two buses"""

    name: Annotated[str, Field(frozen=True, exclude=True)] = ""
    from_to: Annotated[Bus, Field(description="The initial bus", alias="from")]
    to_from: Annotated[Bus, Field(description="The terminal bus", alias="to")]



"""Plugin to add annual carbon cap to the model.

This plugin is only applicable for ReEDs, but could work with similarly arrange data
"""

from argparse import ArgumentParser

import polars as pl
from loguru import logger

from r2x.api import System
from r2x.config_models import PlexosConfig, ReEDSConfig
from r2x.config_scenario import Scenario
from r2x.enums import EmissionType
from r2x.exceptions import R2XModelError
from r2x.models import Emission
from r2x.models.generators import ThermalStandard
from r2x.models.utils import Constraint, ConstraintMap
from r2x.parser.handler import BaseParser
from r2x.units import EmissionRate, ureg
from r2x.utils import get_enum_from_string, validate_string


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--emission-cap",
        type=float,
        help="Emission cap for the solve year in tonnes. Defaults to CO2 or CO2e.",
    )


def update_system(
    config: Scenario,
    system: System,
    parser: BaseParser | None = None,
    emission_cap: float | None = None,
    default_unit: str = "tonne",
) -> System:
    """Apply an emission cap constraint for the system.

    This function adds to the sytem a :class:`~r2x.models.Constraint`object that is used to set the maximum
    emission per year.

    Parameters
    ----------
    config : Scenario
        The scenario configuration.
    parser : BaseParser
        The parser object used for parsing.
    system : System
        The system object to be updated.
    emission_cap : float or None, optional
        The emission cap value. If None, no cap is applied. Default is None.
    default_unit : str, optional
        The default unit for measurement. Default is 'tonne'.

    Notes
    -----
    When summarizing emissions from either fuels or generators, the metric model
    defines one unit in summary (day, week, month, year) as 1000 of the base units,
    whereas the imperial U.S. model uses 2000 units. Thus, if you define a
    constraint on total emissions over a day, week, month, or year, you must
    enter the limit in the appropriate unit. For example, if the production rate
    is in lb/MWh, then an annual constraint would be in short tons, where one
    short ton equals 2000 lbs. For units in kg/MWh and `emission_cap` in metric tons,
    we multiply by 1000 (`Scalar` property in Plexos).
    """
    is_plexos_exporter = isinstance(config.output_config, PlexosConfig)
    if not is_plexos_exporter:
        msg = "Plugin `emission_cap.py` is not compatible with an output model that is not Plexos."
        raise NotImplementedError(msg)

    logger.info("Adding emission cap...")

    if emission_cap is not None:
        logger.debug("Using emission cap value from CLI. Setting emission cap to {}", emission_cap)

    emission_object = EmissionType.CO2  #  This is the default emission object.
    if not any(
        component.emission_type == emission_object
        for component in system.get_supplemental_attributes(Emission)
    ):
        logger.warning("Did not find any emission type to apply emission_cap")
        return system

    is_reeds_parser = parser is not None and isinstance(config.input_config, ReEDSConfig)

    if is_reeds_parser:
        assert parser.data.get("switches") is not None, "Missing switches file from run folder."
        assert isinstance(parser.data["switches"], pl.DataFrame)
        assert parser.data.get("emission_rates") is not None, "Missing emission rates."

        switches = {key: validate_string(value) for key, value in parser.data["switches"].iter_rows()}
        emit_rates = parser.data["emission_rates"]
        emit_rates = emit_rates.with_columns(
            pl.concat_str([pl.col("tech"), pl.col("tech_vintage"), pl.col("region")], separator="_").alias(
                "generator_name"
            )
        )
        any_precombustion = emit_rates["emission_source"].str.contains("precombustion")
        emit_rates = emit_rates.filter(any_precombustion)
        if switches.get("gsw_precombustion") and not emit_rates.is_empty():
            logger.debug("Adding precombustion emission.")
            generator_with_precombustion = emit_rates.select(
                "generator_name", "emission_type", "rate"
            ).unique()
            assert add_precombustion(system, generator_with_precombustion)

    if is_reeds_parser and emission_cap is None:
        emission_object = EmissionType.CO2E if switches["gsw_annualcapco2e"] else EmissionType.CO2
        assert parser.data.get("co2_cap", None) is not None, "co2_cap not found from ReEDS parser"
        emission_cap = parser.data["co2_cap"]["value"].item()

    return set_emission_constraint(system, emission_cap, default_unit, emission_object)


def add_precombustion(system: System, emission_rates: pl.DataFrame) -> bool:
    """Add precombustion emission rates to `Emission` objects.

    This function adds precpmbustion rates to the attributes :class:`~r2x.models.Emission`.

    Parameters
    ----------
    system : System
        The system object to be updated.
    emission_rates : pl.DataFrame
        The precombustion emission_rates

    Returns
    -------
    bool
        True if the addition succeded. False if it failed

    Raises
    ------
    R2XModelError
        If multiple emission_rates of the same type are attached to the component
    """
    applied_rate = False
    for generator_name, emission_type, rate in emission_rates.iter_rows():
        emission_type = get_enum_from_string(emission_type, EmissionType)
        component = system.get_component(ThermalStandard, generator_name)
        attr = system.get_supplemental_attributes_with_component(
            component, Emission, filter_func=lambda attr: attr.emission_type == emission_type
        )
        if not attr:
            logger.trace("`Emission:{}` object not found for {}", emission_type, generator_name)
            continue
        if len(attr) != 1:
            msg = f"Multiple emission of the same type attached to {generator_name}. "
            msg += "Check addition of supplemental attributes."
            raise R2XModelError

        # Extract first element only
        attr = attr[0]

        # Add precombustion emissions
        attr.rate += EmissionRate(rate, attr.rate.units)
        applied_rate = True

    return applied_rate


def set_emission_constraint(
    system: System,
    emission_cap: float | None = None,
    default_unit: str = "tonne",
    emission_object: EmissionType | None = None,
) -> System:
    """Add emissions constraint object to the system."""
    if emission_cap is None:
        logger.warning("Could not set emission cap value. Skipping plugin.")
        return system

    emission_cap = ureg.Quantity(emission_cap, default_unit)

    # All of this are Plexos properties that need to be added
    constraint_properties = {
        "Sense": -1,
        "RHS Year": emission_cap,
        "Scalar": 1e3,
        "Penalty Price": 500,
        emission_object: {"Production Coefficient": 1},
    }
    constraint = Constraint(name=f"Annual_{emission_object}_cap", ext=constraint_properties)

    constraint_map = list(system.get_components(ConstraintMap))
    if len(constraint_map) > 1:
        msg = "Multiple constraint maps are not supported yet."
        raise NotImplementedError(msg)

    if not constraint_map:
        constraint_map = ConstraintMap(name="Constraints")
        system.add_component(constraint_map)

    if isinstance(constraint_map, list):
        constraint_map = constraint_map[0]

    constraint_map.mapping[constraint.name].append(emission_object)

    system.add_component(constraint)
    return system



from infrasys.models import InfraSysBaseModel
from pint import Quantity


class GeoLocation(InfraSysBaseModel):
    Latitude: float
    Longitude: float


class MinMax(InfraSysBaseModel):
    min: float | Quantity
    max: float | Quantity


class UpDown(InfraSysBaseModel):
    up: float | Quantity
    down: float | Quantity


class Complex(InfraSysBaseModel):
    real: float
    imag: float


class InputOutput(InfraSysBaseModel):
    input: float
    output: float


class FromTo_ToFrom(InfraSysBaseModel):  # type: ignore  # noqa: N801
    from_to: float | Quantity
    to_from: float | Quantity


class StartShut(InfraSysBaseModel):
    startup: float | Quantity
    shutdown: float | Quantity


class StartTimeLimits(InfraSysBaseModel):
    hot: float | Quantity
    warm: float | Quantity
    cold: float | Quantity



import pytest
from plexosdb import XMLHandler
from plexosdb.sqlite import PlexosSQLite

from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.exceptions import R2XParserError
from r2x.parser.handler import get_parser_data
from r2x.parser.plexos import PlexosParser
from r2x.models import Generator

DB_NAME = "2-bus_example.xml"
MODEL_NAME = "main_model"


@pytest.fixture
def plexos_scenario(tmp_path, data_folder):
    return Scenario.from_kwargs(
        name="plexos_test",
        input_model="plexos",
        output_model="sienna",
        run_folder=data_folder,
        output_folder=tmp_path,
        model_year=2035,
        user_dict={
            "fmap": {"xml_file": {"fname": DB_NAME, "model_name": MODEL_NAME}},
        },
    )


@pytest.fixture
def pjm_scenario(tmp_path, data_folder):
    return Scenario.from_kwargs(
        name="plexos_test",
        input_model="plexos",
        output_model="sienna",
        run_folder=data_folder / "pjm_2area",
        output_folder=tmp_path,
        model_year=2024,
        user_dict={
            "fmap": {"xml_file": {"fname": "pjm_2area.xml", "model_name": MODEL_NAME}},
        },
    )


@pytest.fixture
def five_bus_variables_scenario(tmp_path, data_folder):
    return Scenario.from_kwargs(
        name="plexos_test",
        input_model="plexos",
        output_model="sienna",
        run_folder=data_folder / "5_bus_system_variables",
        output_folder=tmp_path,
        model_year=2035,
        fmap={"xml_file": {"fname": "5_bus_system_variables.xml", "model": "Base"}},
        user_dict={
            "fmap": {"xml_file": {"fname": "5_bus_system_variables.xml", "model": "Base"}},
        },
    )


@pytest.fixture
def plexos_parser_instance(plexos_scenario):
    plexos_device_map = {"SolarPV_01": "RenewableFix", "ThermalCC": "ThermalStandard"}
    plexos_scenario.input_config.defaults["plexos_device_map"] = plexos_device_map
    return get_parser_data(plexos_scenario, parser_class=PlexosParser)


def test_plexos_parser_instance(plexos_parser_instance):
    assert isinstance(plexos_parser_instance, PlexosParser)
    assert len(plexos_parser_instance.data) == 1  # Plexos parser just parses a single file
    assert isinstance(plexos_parser_instance.data["xml_file"], XMLHandler)
    assert isinstance(plexos_parser_instance.db, PlexosSQLite)


@pytest.mark.skip(reason="We need a better test model")
def test_build_system(plexos_parser_instance):
    system = plexos_parser_instance.build_system()
    assert isinstance(system, System)


def test_parser_system(pjm_scenario):
    plexos_category_map = {
        "thermal": {"fuel": "NATURAL_GAS", "type": "CC"},
        "solar": {"fuel": None, "type": "PVe"},
        "wind": {"fuel": None, "type": "WT"},
    }
    pjm_scenario.input_config.model_name = "model_2012"

    with pytest.raises(R2XParserError):
        parser = get_parser_data(pjm_scenario, parser_class=PlexosParser)

    pjm_scenario.input_config.defaults["plexos_category_map"] = plexos_category_map

    parser = get_parser_data(pjm_scenario, parser_class=PlexosParser)
    system = parser.build_system()
    assert isinstance(system, System)

    # PJM system has 48 components
    total_components = sum(1 for _ in system.iter_all_components())
    assert total_components == 48


def test_variable_parsing(five_bus_variables_scenario):
    plexos_category_map = {
        "thermal": {"fuel": "NATURAL_GAS", "type": "CC"},
        "solar": {"fuel": None, "type": "WT"},
        "wind": {"fuel": None, "type": "PV"},
    }
    five_bus_variables_scenario.input_config.model_name = "Base"

    five_bus_variables_scenario.input_config.defaults["plexos_category_map"] = plexos_category_map

    parser = get_parser_data(five_bus_variables_scenario, parser_class=PlexosParser)
    system = parser.build_system()
    assert isinstance(system, System)

    record_ts = {}
    for component in system.get_components(Generator):
        if not system.has_time_series(component):
            continue
        ts_metadata = system.get_time_series(component, "max_active_power")
        record_ts[component.name] = ts_metadata

    assert sum(record_ts["SolarPV1"].data.tolist()) == 4224.0
    assert "SolarPV2" not in record_ts



import h5py
import numpy as np
import pandas as pd
import pytest
from _pytest.logging import LogCaptureFixture
from loguru import logger

WEATHER_YEARS = 7


@pytest.fixture
def caplog(caplog: LogCaptureFixture):
    handler_id = logger.add(
        caplog.handler,
        format="{message}",
        level=0,
        filter=lambda record: record["level"].no >= caplog.handler.level,
        enqueue=False,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    logger.remove(handler_id)


@pytest.fixture(scope="function")
def pandas_h5_weather_year(tmp_path):
    fpath = tmp_path / "pandas.h5"
    index = pd.Index(range(WEATHER_YEARS * 8760), name="index")
    data = {f"upv_{number}": np.random.rand(WEATHER_YEARS * 8760) for number in range(3)}
    pandas_data = pd.DataFrame(index=index, data=data)
    pandas_data.to_hdf(fpath, key="df", mode="w", format="fixed")
    return fpath


@pytest.fixture(scope="function")
def pandas_h5_solve_year_and_weather_year(tmp_path):
    fpath = tmp_path / "pandas.h5"

    unique_year_sorted = sorted(set(range(2030, 2050, 10)))
    unique_datetime_sorted = sorted(set(range(WEATHER_YEARS * 8760)))
    multilevel_index = pd.MultiIndex.from_product(
        [unique_year_sorted, unique_datetime_sorted], names=["year", "datetime"]
    )
    data_size = len(multilevel_index)
    data = {f"upv_{number}": np.random.rand(data_size) for number in range(3)}
    pandas_data = pd.DataFrame(index=multilevel_index, data=data)
    pandas_data.to_hdf(fpath, key="df", mode="w", format="fixed")
    return fpath


@pytest.fixture(scope="function")
def h5_without_index_names(tmp_path):
    fpath = tmp_path / "h5_no_index.h5"
    with h5py.File(fpath, "w") as f:
        f.create_dataset("index_0", data=[0])
        f.create_dataset("index_1", data=[0])
        f.create_dataset("columns", data=[0])
        f.create_dataset("index_names", data=[0])
        f.create_dataset("data", data=[0])
    return fpath


@pytest.fixture(scope="function")
def h5_with_index_names_no_datetime(tmp_path):
    fpath = tmp_path / "h5_no_index.h5"
    unique_year_sorted = sorted(set(range(2030, 2050, 10)))
    unique_datetime_sorted = sorted(set(range(WEATHER_YEARS * 8760)))
    data_keys = ["upv_1", "upv_2", "upv_3"]
    data = np.random.rand(WEATHER_YEARS * 8760, len(data_keys))
    with h5py.File(fpath, "w") as f:
        f.create_dataset("index_year", data=unique_year_sorted)
        f.create_dataset("index_datetime", data=unique_datetime_sorted)
        f.create_dataset("columns", data=data_keys)
        f.create_dataset("index_names", data=["year", "datetime"])
        f.create_dataset("data", data=data)
    return fpath



"""Supplemental attributes models."""

from typing import Annotated

from infrasys import SupplementalAttribute
from pydantic import Field

from r2x.enums import EmissionType
from r2x.models.named_tuples import GeoLocation
from r2x.units import EmissionRate


class GeographicInfo(SupplementalAttribute):
    """Supplemental attribute that capture location."""

    geo_json: GeoLocation

    @classmethod
    def example(cls) -> "GeographicInfo":
        return GeographicInfo(geo_json=GeoLocation(Latitude=10.5, Longitude=-100))


class Emission(SupplementalAttribute):
    """Class representing an emission object that is attached to generators."""

    rate: Annotated[EmissionRate, Field(description="Amount of emission produced in kg/MWh.")]
    emission_type: Annotated[EmissionType, Field(description="Type of emission. E.g., CO2, NOx.")]

    @classmethod
    def example(cls) -> "Emission":
        return Emission(
            rate=EmissionRate(105, "kg/MWh"),
            emission_type=EmissionType.CO2,
        )



"""Umbrella API for R2X model."""

import importlib
import inspect
import shutil
import sys
from importlib.resources import files
from pathlib import Path

from loguru import logger

from r2x.exporter.handler import get_exporter

from .api import System
from .config_scenario import Scenario, get_scenario_configuration
from .exporter import exporter_list
from .parser import parser_list
from .parser.handler import BaseParser, get_parser_data
from .upgrader import upgrade_handler
from .utils import (
    DEFAULT_PLUGIN_PATH,
)


def run_parser(config: Scenario, **kwargs):
    """Call get parser for parser selected.

    The parser gets selected based on `config.input_model`.

    Parameters
    ----------
    config
        Scenario
    upgrader
        Function to upgrade the inputs

    Other Parameters
    ----------------
    kwargs
        Additional key arguments to the parser.

    Raises
    ------
    KeyError
        If parser is not found on parser_list
    """
    # At some point we will read the ReEDS tag and upgrade accordingly.
    # reeds_meta = get_file(fname="meta.csv", config=config).iloc[:1]
    assert config.run_folder
    assert config.input_model
    if getattr(config, "upgrade", None):
        upgrade_handler(config.run_folder)

    # Initialize parser
    parser_class = parser_list.get(config.input_model)
    if not parser_class:
        raise KeyError(f"Parser for {config.input_model} not found")

    parser = get_parser_data(config, parser_class, **kwargs)
    system = parser.build_system()

    assert system is not None, "System failed to create"

    return system, parser


def run_plugins(config: Scenario, parser: BaseParser, system: System) -> System:
    """Run selected plugins.

    Parameters
    ----------
    config
        Configuration manager
    system
        System

    Returns
    -------
        System
    """
    if not config.plugins:
        return system

    logger.info("Running the following plugins: {}", config.plugins)
    for plugin in config.plugins:
        module = importlib.import_module(f".{plugin}", DEFAULT_PLUGIN_PATH)
        if hasattr(module, "update_system"):
            plugin_required_args = inspect.getfullargspec(module.update_system).args
            plugin_config_args = {
                key: value for key, value in config.__dict__.items() if key in plugin_required_args
            }
            system = module.update_system(config=config, parser=parser, system=system, **plugin_config_args)
    return system


def run_exporter(config: Scenario, system: System) -> None:
    """Create exporter model."""
    assert config.output_model
    exporter_class = exporter_list.get(config.output_model, None)
    if not exporter_class:
        raise KeyError(f"Exporter for {config.output_model} not found")

    get_exporter(config, system, exporter_class)


def run_single_scenario(scenario: Scenario, **kwargs) -> None:
    """Run translation process."""
    logger.info("Running {}", scenario.name)

    if scenario.input_model == "infrasys":
        fname = f"{scenario.run_folder}/{scenario.name}.json"
        system = System.from_json(filename=fname, **kwargs)
    else:
        system, parser = run_parser(scenario, **kwargs)
        system = run_plugins(config=scenario, parser=parser, system=system)

    if getattr(scenario, "inspect", None):
        from IPython import embed

        embed()
        sys.exit(0)

    # Serialize/deserialize the system
    output_fpath = str(scenario.output_folder) + f"/{scenario.name}.json"
    if getattr(scenario, "save", None):
        logger.info("Serialize system to {}", output_fpath)
        system.to_json(output_fpath, overwrite=True)
        system = System.from_json(output_fpath)

    if scenario.output_model == "infrasys":
        logger.info("Serialize system to {}", output_fpath)
        system.to_json(output_fpath, overwrite=True)
        return
    run_exporter(config=scenario, system=system)
    return


def run(cli_args: dict, user_dict: dict | None = None) -> None:
    """Run a translation.

    This function takes the `cli_args` dictionary to create the configuratio and run the translation process.
    If the user specifies multiple scenarios, this function run the translation sequentially with
    no specific order.

    Parameters
    ----------
    cli_args
        Arguments for constructing the scenario.
    user_dict
        Optional, User custom configuration.

    See Also
    --------
        get_config: Get configuration from arguments
        run_single_scenario: Run a single translation scenario

    Notes
    -----
    Currently the scenario should only have a single year to run.
    """
    config_mgr = get_scenario_configuration(cli_args=cli_args, user_dict=user_dict)
    logger.info("Running {} scenarios", len(config_mgr))
    for _, scenario in config_mgr.scenarios.items():
        # NOTE: We can pass multiple years from the CLI. In those cases we want to raise not implemented.
        if hasattr(scenario, "input_config") and isinstance(
            getattr(scenario.input_config, "solve_year", None), list
        ):
            msg = "Multi year runs from the CLI is not yet supported. Use scenarios instead."
            raise NotImplementedError(msg)
        run_single_scenario(scenario)
    return


def init(cli_args: dict) -> None:
    """Copy the default configuration file on the path that the user request.

    If the user does not provide a path, the current path is used.

    Parameters
    ----------
    cli_args
        Arguments from the CLI
    """
    logger.debug("Running init command")
    path = Path(cli_args["path"])

    with files("r2x.defaults") as package_path:  # type: ignore
        file_path = package_path / "user_dict.yaml"
        shutil.copy(file_path, Path(path) / "user_dict.yaml")
    return



import numpy
import polars
import pytest
from pint import Quantity

from r2x.api import System
from r2x.config_models import SiennaConfig
from r2x.config_scenario import Scenario
from r2x.enums import EmissionType
from r2x.exceptions import R2XModelError
from r2x.models import Emission
from r2x.models.generators import ThermalStandard
from r2x.models.utils import Constraint, ConstraintMap
from r2x.plugins.emission_cap import add_precombustion, update_system
from r2x.runner import run_parser, run_plugins
from r2x.units import EmissionRate


@pytest.fixture(scope="function")
def test_system(reeds_data_folder, tmp_path):
    config = Scenario.from_kwargs(
        name="ReEDS-TestSystem",
        run_folder=reeds_data_folder,
        output_folder=tmp_path,
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
        plugins=["emission_cap"],
    )
    system, parser = run_parser(config)
    return config, system, parser


def test_update_system_default(test_system):
    config, system, parser = test_system

    new_system = update_system(config=config, parser=parser, system=system)
    assert isinstance(new_system, System)

    constraint = next(iter(new_system.get_components(Constraint)))
    assert constraint
    assert isinstance(constraint, Constraint)
    assert constraint.ext is not None
    assert constraint.ext.get("RHS Year") is not None
    assert isinstance(constraint.ext["RHS Year"], Quantity)
    assert numpy.isclose(constraint.ext["RHS Year"].magnitude, 1.14e9, rtol=1e-02)

    constraint_map = next(iter(new_system.get_components(ConstraintMap)))
    assert constraint_map
    assert isinstance(constraint_map, ConstraintMap)
    assert EmissionType.CO2 in constraint_map.mapping[constraint.name]

    system, parser = run_parser(config)
    new_system = update_system(config=config, parser=parser, system=system, emission_cap=0.0)
    assert isinstance(new_system, System)
    constraint = next(iter(new_system.get_components(Constraint)))
    assert constraint
    assert isinstance(constraint, Constraint)
    assert constraint.ext is not None
    assert constraint.ext.get("RHS Year") is not None
    assert isinstance(constraint.ext["RHS Year"], Quantity)
    assert constraint.ext["RHS Year"].magnitude == 0

    # Test invalid models
    config.output_config = SiennaConfig()
    with pytest.raises(NotImplementedError):
        _ = update_system(config=config, parser=parser, system=system)


def test_no_emission(caplog, infrasys_test_system):
    config = Scenario.from_kwargs(
        name="Pacific",
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
        emission_cap=0.0,
        plugins=["emission_cap"],
    )

    _ = update_system(config=config, system=infrasys_test_system)
    assert "Did not find any emission" in caplog.text


def test_emission_but_no_cap(caplog):
    system = System(name="Test")
    gen1 = ThermalStandard.example()
    emission = Emission(rate=10, emission_type=EmissionType.CO2)
    system.add_supplemental_attribute(gen1, emission)
    config = Scenario.from_kwargs(
        name="Pacific",
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
        plugins=["emission_cap"],
    )

    _ = update_system(config=config, system=system)
    assert "Could not set emission cap value" in caplog.text


def test_multiple_constraint_maps():
    system = System(name="Test")
    gen1 = ThermalStandard.example()
    emission = Emission(rate=10, emission_type=EmissionType.CO2)
    constraint_map = ConstraintMap(name="Constraints")
    constraint_map2 = ConstraintMap(name="Constraints")
    system.add_components(constraint_map, constraint_map2)
    system.add_supplemental_attribute(gen1, emission)
    config = Scenario.from_kwargs(
        name="Pacific",
        input_model="reeds-US",
        output_model="plexos",
        solve_year=2035,
        weather_year=2012,
        plugins=["emission_cap"],
    )

    with pytest.raises(NotImplementedError):
        _ = update_system(config=config, system=system, emission_cap=0.0)


def test_update_system_using_cli(test_system):
    config, system, parser = test_system
    config.emission_cap = 0.0
    new_system = run_plugins(config=config, parser=parser, system=system)
    assert isinstance(new_system, System)
    constraint = next(iter(new_system.get_components(Constraint)))
    assert constraint
    assert isinstance(constraint, Constraint)
    assert constraint.ext is not None
    assert constraint.ext.get("RHS Year") is not None
    assert isinstance(constraint.ext["RHS Year"], Quantity)
    assert constraint.ext["RHS Year"].magnitude == 0


def test_emission_source(test_system, caplog):
    config, system, parser = test_system

    # Manually adding the switch
    # NOTE: We might need to modify this if we change the parsing of this file
    # https://github.com/NREL/R2X/issues/177
    adding_precombustion_switch = polars.DataFrame([{"aws": "gsw_precombustion", "0": "true"}])
    parser.data["switches"] = polars.concat([parser.data["switches"], adding_precombustion_switch])
    # Adding precombustion to the first generator
    adding_precombustion_entry = polars.DataFrame(
        [
            {
                "emission_type": "so2",
                "emission_source": "precombustion",
                "tech": "biopower",
                "region": "p10",
                "year": 2035,
                "tech_vintage": "init-2",
                "rate": -0.000642,  # Value to make it 0 for testing purposes
            }
        ]
    )
    parser.data["emission_rates"] = polars.concat([parser.data["emission_rates"], adding_precombustion_entry])
    new_system = update_system(config=config, parser=parser, system=system)
    assert "precombustion" in caplog.text

    # Check operation was correct
    generator_component = new_system.get_component(ThermalStandard, name="biopower_init-2_p10")
    assert generator_component
    attribute = new_system.get_supplemental_attributes_with_component(
        generator_component, Emission, filter_func=lambda x: x.emission_type == "SO2"
    )
    assert attribute
    assert len(attribute) == 1
    assert attribute[0].rate == 0.0


def test_add_precombustion(caplog):
    system = System(auto_add_composed_components=True)
    gen1 = ThermalStandard.example()
    system.add_component(gen1)
    emission_rates = polars.DataFrame(
        [{"generator_name": "ThermalStandard", "emission_type": "CO2", "rate": 10}]
    )
    assert not add_precombustion(system, emission_rates)
    assert "object not found" in caplog.text

    emission = Emission(rate=EmissionRate(10, "kg/MWh"), emission_type=EmissionType.CO2)
    system.add_supplemental_attribute(gen1, emission)
    assert add_precombustion(system, emission_rates)
    assert emission.rate.magnitude == 20

    emission_02 = Emission(rate=EmissionRate(10, "kg/MWh"), emission_type=EmissionType.CO2)
    system.add_supplemental_attribute(gen1, emission_02)
    with pytest.raises(R2XModelError):
        add_precombustion(system, emission_rates)



"""Helper functions for upgrader."""

import datetime
import inspect
from collections.abc import Callable
from pathlib import Path

import h5py
import numpy as np
import pandas as pd
from loguru import logger

from r2x.utils import get_timeindex, validate_string


def add_datetime_index(fpath, datetime_column: str = "index_datetime") -> bool:
    """Add datetime index for h5 file w/o it."""
    time_index = get_timeindex()
    time_index = time_index.to_series().apply(datetime.datetime.isoformat).reset_index(drop=True)
    time_index = np.char.encode(time_index.to_numpy().astype(str), "utf-8")

    with h5py.File(fpath, "r+") as f:
        year_multiplier = len(set(f["index_year"][:])) if "index_year" in f.keys() else 1
        time_index = np.tile(time_index, year_multiplier)
        del f[datetime_column]
        f.create_dataset(datetime_column, data=time_index, dtype="S30")
    return True


def rename_index_names_from_h5(fpath) -> bool:
    """Rename index for h5 file."""
    multi_index = {"index_0", "index_1"}
    single_index = {"index_0"}

    with h5py.File(fpath, "r+") as f:
        keys = set(f.keys())

        match keys:
            case set() if multi_index.issubset(keys):
                column_mapping = {"index_0": "index_year", "index_1": "index_datetime"}
            case set() if single_index.issubset(keys):
                column_mapping = {"index_0": "index_datetime"}
        for old_name, new_name in column_mapping.items():
            f.move(old_name, new_name)
    return True


def pandas_to_h5py(pandas_dataframe: pd.DataFrame, output_fpath: Path, compression_opts: int = 4) -> bool:
    """Convert a pandas dataframe to an h5py."""
    logger.debug("Converting pandas style H5 {} to h5py compatible", output_fpath)
    timeindex = get_timeindex()
    timeindex = timeindex.to_series().apply(datetime.datetime.isoformat).reset_index(drop=True)
    with h5py.File(output_fpath, "w") as f:
        match pandas_dataframe.index:
            case pd.MultiIndex():
                for level in pandas_dataframe.index.levels:
                    if (len(level) == 7 * 8760) and (isinstance(level.to_numpy()[0], np.int64)):
                        assert len(level) == len(timeindex), (
                            f"H5 file {output_fpath} has more weather year data."
                        )
                        f.create_dataset("index_datetime", data=timeindex.str.encode("utf-8"), dtype="S30")
                    else:
                        f.create_dataset(f"index_{level.name}", data=level.values, dtype=level.dtype)

            case pd.Index():
                f.create_dataset("index_datetime", data=timeindex.str.encode("utf-8"), dtype="S30")

        index_names = pd.Index(pandas_dataframe.index.names)
        f.create_dataset("index_names", data=index_names, dtype=f"S{index_names.map(len).max()}")

        f.create_dataset(
            "columns", data=pandas_dataframe.columns, dtype=f"S{pandas_dataframe.columns.map(len).max()}"
        )

        if len(pandas_dataframe.dtypes.unique()) > 1:
            msg = f"Multiple data types detected in {output_fpath.name}, "
            msg += "unclear which one to use for re-saving h5."
            raise Exception(msg)
        pandas_dataframe = pandas_dataframe.astype(np.float32)
        dftype_out = pandas_dataframe.dtypes.unique()[0]
        f.create_dataset(
            "data",
            data=pandas_dataframe.values,
            dtype=dftype_out,
            compression="gzip",
            compression_opts=compression_opts,
        )
    return True


def get_function_arguments(argument_input: dict, function: Callable) -> dict:
    """Get arguments to pass to a function based on its signature.

    This function processes the `argument_input` and returns a dictionary of argument
    values that are valid for the given `function`, using the function's signature
    as a filter. String values are validated, nested dictionaries are flattened,
    and only the valid argument keys (as defined in the function signature) are included.

    Parameters
    ----------
    data_dict : dict
        A dictionary containing potential argument values, which may include
        strings, dictionaries, and other types of data.

    function : str
        The name of the function whose signature is used to filter the arguments.

    Returns
    -------
    dict
        A dictionary of filtered arguments that match the function's signature.
        Only arguments that exist in the function's signature will be included.

    Example
    -------
    >>> def example_function(a, b, c=None):
    >>>     pass
    >>> data = {"a": 1, "b": 2, "c": 3, "extra": 4}
    >>> prepare_function_arguments(data, "example_function")
    {'a': 1, 'b': 2, 'c': 3}
    """
    arguments = {}
    for key, value in argument_input.items():
        if isinstance(value, str):
            value = validate_string(value)
        if isinstance(value, dict):
            for sub_key, sub_value in value.items():
                arguments[sub_key] = sub_value
        else:
            arguments[key] = value

    return {key: value for key, value in arguments.items() if key in inspect.getfullargspec(function).args}



"""Custom exceptions for R2X."""
# ruff: noqa: D101


class R2XDuplicateUUIDError(Exception):
    pass


class R2XDuplicateNameError(Exception):
    pass


class R2XNotFoundError(Exception):
    pass


class R2XMultlipleElementsError(Exception):
    pass


class R2XModelError(Exception):
    pass


class R2XMultipleFilesError(Exception):
    pass


class R2XParserError(Exception):
    pass


class R2XFieldRemovalError(Exception):
    pass



"""Plugin to dissaggregate and aggregate generators.

This plugin breaks apart generators that are to big in conmparisson with the
WECC database. If the generator is to small after the breakup than the capacity
threshold variable, we drop the genrator entirely.
"""

# System packages
import re
from argparse import ArgumentParser

import numpy as np
import pandas as pd
from infrasys.base_quantity import BaseQuantity

# Local imports
from loguru import logger

from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.models import Emission, Generator
from r2x.parser.handler import BaseParser
from r2x.units import ActivePower, ureg
from r2x.utils import read_json

# Constants
CAPACITY_THRESHOLD = 5  # MW
PROPERTIES_TO_BREAK = [
    "ramp_up",
    "ramp_down",
    "min_rated_capacity",
    "startup_cost",
    "pump_load",
    "storage_capacity",
]


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--capacity-threshold",
        type=int,
        help="Capacity threshold for aggregating generators.",
        default=CAPACITY_THRESHOLD,
    )


def update_system(
    config: Scenario,
    parser: BaseParser,
    system: System,
    pcm_defaults_fpath: str | None = None,
    capacity_threshold: int = CAPACITY_THRESHOLD,
) -> System:
    """Break apart large generators based on average capacity.

    This function updates and overrides the data objects from the translator to
    be passed to the exporter.

    Args:
        translator: r2x.Translator object,
        *args: additional arguments that can be passed,
        **kwargs: additiona arguments.
    """
    logger.info("Dividing generators into average size generators")
    assert config.input_config
    if pcm_defaults_fpath is None:
        logger.debug("Using {}", config.input_config.defaults["pcm_defaults_fpath"])
        pcm_defaults = read_json(config.input_config.defaults["pcm_defaults_fpath"])
    else:
        logger.debug("Using custom defaults from {}", pcm_defaults_fpath)
        pcm_defaults: dict = read_json(pcm_defaults_fpath)

    reference_generators = (
        pd.DataFrame.from_dict(pcm_defaults)
        .transpose()
        .reset_index()
        .rename(
            columns={
                "index": "tech",
            }
        )
        .set_index("tech")
        .replace({np.nan: None})
        .to_dict(orient="index")
    )
    non_break_techs = config.input_config.defaults.get("non_break_techs", [])

    return break_generators(system, reference_generators, capacity_threshold, non_break_techs)


def break_generators(  # noqa: C901
    system: System,
    reference_generators: dict[str, dict],
    capacity_threshold: int = CAPACITY_THRESHOLD,
    non_break_techs: list[str] | None = None,
    break_category: str = "category",
) -> System:
    """Break component generator into smaller units."""
    if non_break_techs:
        regex_pattern = f"^(?!{'|'.join(non_break_techs)})."  # Oh yes.
    else:
        regex_pattern = ".*"

    capacity_dropped = 0  # count capacity dropped
    for component in system.get_components(Generator, filter_func=lambda x: re.search(regex_pattern, x.name)):
        if not (tech := getattr(component, break_category, None)):
            logger.trace("Skipping component {} with missing category", component.label)
            continue
        logger.trace("Breaking {}", component.label)

        if not (reference_tech := reference_generators.get(tech)):
            logger.trace("{} not found in reference_generators", tech)
            continue
        if not (avg_capacity := reference_tech.get("avg_capacity_MW", None)):
            continue
        logger.trace("Average_capacity: {}", avg_capacity)
        reference_base_power = (
            component.active_power.magnitude
            if isinstance(component.active_power, BaseQuantity)
            else component.active_power
        )
        no_splits = int(reference_base_power // avg_capacity)
        remainder = reference_base_power % avg_capacity

        # If there is not splits, we skip the generator
        if no_splits <= 1:
            continue
        split_no = 1
        logger.trace(
            "Breaking generator {} with active_power {} into {} generators of {} capacity",
            component.name,
            reference_base_power,
            no_splits,
            avg_capacity,
        )
        for _ in range(no_splits):
            component_name = component.name + f"_{split_no:02}"
            new_component = system.copy_component(component, name=component_name, attach=True)
            new_base_power = (
                ActivePower(avg_capacity, component.active_power.units)
                if isinstance(component.active_power, BaseQuantity)
                else avg_capacity * ureg.MW
            )
            new_component.active_power = new_base_power
            proportion = (
                avg_capacity / reference_base_power
            )  # Required to recalculate properties that depend on active_power
            for property in PROPERTIES_TO_BREAK:
                if attr := getattr(new_component, property, None):
                    new_component.ext[f"{property}_original"] = attr
                    setattr(new_component, property, attr * proportion)
            new_component.ext["original_capacity"] = component.active_power
            new_component.ext["original_name"] = component.name
            new_component.ext["broken"] = True

            for attribute in system.get_supplemental_attributes_with_component(component, Emission):
                system.add_supplemental_attribute(new_component, attribute)
            if system.has_time_series(component):
                logger.trace(
                    "Component {} has time series attached to it. Copying first one", component.label
                )
                ts = system.get_time_series(component)
                system.add_time_series(ts, new_component)
            split_no += 1

        if remainder > capacity_threshold:
            component_name = component.name + f"_{split_no:02}"
            new_component = system.copy_component(component, name=component_name, attach=True)
            new_component.active_power = remainder * ureg.MW
            proportion = (
                remainder / reference_base_power
            )  # Required to recalculate properties that depend on active_power
            for property in PROPERTIES_TO_BREAK:
                if attr := getattr(new_component, property, None):
                    new_component.ext[f"{property}_original"] = attr
                    setattr(new_component, property, attr * proportion)
            new_component.ext["original_capacity"] = component.active_power
            new_component.ext["original_name"] = component.name
            new_component.ext["broken"] = True
            for attribute in system.get_supplemental_attributes_with_component(component, Emission):
                system.add_supplemental_attribute(new_component, attribute)

            if system.has_time_series(component):
                logger.trace(
                    "Component {} has time series attached to it. Copying first one", component.label
                )
                ts = system.get_time_series(component)
                system.add_time_series(ts, new_component)
        else:
            capacity_dropped = capacity_dropped + remainder
            logger.debug("Dropped {} capacity for {}", remainder, component.name)

        # Finally remove the component and emissions from original component
        system.remove_component(component)
        for attribute in system.get_supplemental_attributes_with_component(component):
            system.remove_supplemental_attribute_from_component(component, attribute)

    logger.info("Total capacity dropped {} MW", capacity_dropped)
    return system



import pytest
from r2x.runner import init, run


def test_runner(tmp_path, reeds_data_folder):
    cli_input = {
        "name": "Test",
        "weather_year": 2012,
        "solve_year": [2050],
        "input_model": "reeds-US",
        "output_model": "sienna",
        "output_folder": str(tmp_path),
        "feature_flags": {"cool-feature": True},
        "run_folder": reeds_data_folder,
    }

    _ = run(cli_input, {})


def test_runner_exceptions(tmp_path, reeds_data_folder):
    cli_input = {
        "name": "Test",
        "weather_year": 2015,
        "solve_year": [2055, 2050],
        "input_model": "reeds-US",
        "output_model": "sienna",
        "output_folder": str(tmp_path),
        "feature_flags": {"cool-feature": True},
        "run_folder": reeds_data_folder,
    }

    with pytest.raises(NotImplementedError):
        _ = run(cli_input, {})


def test_runner_serialization(tmp_path, reeds_data_folder):
    cli_input = {
        "name": "Test",
        "weather_year": 2012,
        "solve_year": [2050],
        "input_model": "reeds-US",
        "output_model": "sienna",
        "output_folder": str(tmp_path),
        "feature_flags": {"cool-feature": True},
        "run_folder": reeds_data_folder,
        "save": True,
    }

    _ = run(cli_input, {})
    assert (tmp_path / f"{cli_input['name']}.json").exists()


def test_init(tmp_path):
    cli_input = {"path": str(tmp_path)}
    _ = init(cli_input)
    assert (tmp_path / "user_dict.yaml").exists()



"""Models for generator devices."""

from typing import Annotated, Any

from pint import Quantity
from pydantic import Field, NonNegativeFloat, field_serializer

from r2x.enums import PrimeMoversType, StorageTechs, ThermalFuels
from r2x.models.core import Device
from r2x.models.costs import HydroGenerationCost, RenewableGenerationCost, StorageCost, ThermalGenerationCost
from r2x.models.load import PowerLoad
from r2x.models.named_tuples import InputOutput, MinMax, StartShut, StartTimeLimits, UpDown
from r2x.models.topology import ACBus
from r2x.units import (
    ActivePower,
    ApparentPower,
    Energy,
    Percentage,
    PowerRate,
    Time,
    VOMPrice,
    ureg,
)


class Generator(Device):
    """Abstract generator class."""

    bus: Annotated[ACBus, Field(description="Bus where the generator is connected.")] | None = None
    rating: Annotated[
        ApparentPower | None,
        Field(ge=0, description="Maximum output power rating of the unit (MVA)."),
    ] = ApparentPower(1, "MVA")
    active_power: Annotated[
        ActivePower,
        Field(
            description=(
                "Initial active power set point of the unit in MW. For power flow, this is the steady "
                "state operating point of the system."
            ),
        ),
    ] = ActivePower(0.0, "MW")
    reactive_power: Annotated[
        ApparentPower | None,
        Field(
            description=(
                "Reactive power set point of the unit in MW. For power flow, this is the steady "
                "state operating point of the system."
            ),
        ),
    ] = ApparentPower(0.0, "MVA")
    base_mva: float = 1
    base_power: Annotated[
        ApparentPower | None,
        Field(
            gt=0,
            description="Base power of the unit (MVA) for per unitization.",
        ),
    ] = None
    must_run: Annotated[int | None, Field(description="If we need to force the dispatch of the device.")] = (
        None
    )
    vom_price: Annotated[VOMPrice, Field(description="Variable operational price $/MWh.")] | None = None
    prime_mover_type: (
        Annotated[PrimeMoversType, Field(description="Prime mover technology according to EIA 923.")] | None
    ) = None
    unit_type: Annotated[
        PrimeMoversType | None, Field(description="Prime mover technology according to EIA 923.")
    ] = None
    min_rated_capacity: Annotated[ActivePower, Field(description="Minimum rated power generation.")] = (
        0 * ureg.MW
    )
    ramp_up: (
        Annotated[
            PowerRate,
            Field(description="Ramping rate on the positve direction."),
        ]
        | None
    ) = None
    ramp_down: (
        Annotated[
            PowerRate,
            Field(description="Ramping rate on the negative direction."),
        ]
        | None
    ) = None
    min_up_time: (
        Annotated[
            Time,
            Field(ge=0, description="Minimum up time in hours for UC decision."),
        ]
        | None
    ) = None
    min_down_time: (
        Annotated[
            Time,
            Field(ge=0, description="Minimum down time in hours for UC decision."),
        ]
        | None
    ) = None
    mean_time_to_repair: (
        Annotated[
            Time,
            Field(gt=0, description="Total hours to repair after outage occur."),
        ]
        | None
    ) = None
    forced_outage_rate: (
        Annotated[
            Percentage,
            Field(description="Expected level of unplanned outages in percent."),
        ]
        | None
    ) = None
    planned_outage_rate: (
        Annotated[
            Percentage,
            Field(description="Expected level of planned outages in percent."),
        ]
        | None
    ) = None
    startup_cost: Annotated[NonNegativeFloat, Field(description="Cost in $ of starting a unit.")] | None = (
        None
    )
    shutdown_cost: (
        Annotated[NonNegativeFloat, Field(description="Cost in $ of shuting down a unit.")] | None
    ) = None
    active_power_limits: Annotated[
        MinMax | None, Field(description="Maximum output power rating of the unit (MVA).")
    ] = None
    reactive_power_limits: Annotated[
        MinMax | None, Field(description="Maximum output power rating of the unit (MVA).")
    ] = None

    @field_serializer("active_power_limits")
    def serialize_active_power_limits(self, min_max: MinMax | dict | None) -> dict[str, Any] | None:
        if min_max is None:
            return None
        if not isinstance(min_max, MinMax):
            min_max = MinMax(**min_max)
        if min_max is not None:
            return {
                "min": min_max.min.magnitude if isinstance(min_max.min, Quantity) else min_max.min,
                "max": min_max.max.magnitude if isinstance(min_max.max, Quantity) else min_max.max,
            }


class RenewableGen(Generator):
    """Abstract class for renewable generators."""


class RenewableDispatch(RenewableGen):
    """Curtailable renewable generator.

    This type of generator have a hourly capacity factor profile.
    """

    power_factor: float = 1.0
    operation_cost: RenewableGenerationCost | None = None


class RenewableNonDispatch(RenewableGen):
    """Non-curtailable renewable generator.

    Renewable technologies w/o operational cost.
    """

    power_factor: float = 1.0


class HydroGen(Generator):
    """Hydroelectric generator."""


class HydroDispatch(HydroGen):
    """Class representing flexible hydro generators."""

    operation_cost: HydroGenerationCost | None = None
    ramp_up: (
        Annotated[
            PowerRate,
            Field(ge=0, description="Ramping rate on the positve direction."),
        ]
        | None
    ) = None
    ramp_down: (
        Annotated[
            PowerRate,
            Field(ge=0, description="Ramping rate on the negative direction."),
        ]
        | None
    ) = None


class HydroFix(HydroGen):
    """Class representing unflexible hydro."""


class HydroEnergyReservoir(HydroGen):
    """Class representing hydro system with reservoirs."""

    ramp_limits: UpDown | None = None
    time_limits: UpDown | None = None
    operation_cost: HydroGenerationCost | None = None
    inflow: float | None = None
    initial_energy: (
        Annotated[NonNegativeFloat, Field(description="Initial water volume or percentage.")] | None
    ) = 0
    initial_storage: float | None = None
    storage_capacity: (
        Annotated[
            Energy,
            Field(description="Total water volume or percentage."),
        ]
        | None
    ) = None
    min_storage_capacity: (
        Annotated[
            Energy,
            Field(description="Minimum water volume or percentage."),
        ]
        | None
    ) = None
    storage_target: (
        Annotated[
            Energy,
            Field(description="Maximum energy limit."),
        ]
        | None
    ) = None


class HydroPumpedStorage(HydroGen):
    """Class representing pumped hydro generators."""

    operation_cost: HydroGenerationCost | StorageCost | None = None
    storage_duration: (
        Annotated[
            Time,
            Field(description="Storage duration in hours."),
        ]
        | None
    ) = None
    initial_volume: (
        Annotated[Energy, Field(gt=0, description="Initial water volume or percentage.")] | None
    ) = None
    storage_capacity: Annotated[
        UpDown,
        Field(description="Total water volume or percentage."),
    ]
    min_storage_capacity: (
        Annotated[
            Energy,
            Field(description="Minimum water volume or percentage."),
        ]
        | None
    ) = None
    pump_efficiency: Annotated[Percentage, Field(ge=0, le=1, description="Pumping efficiency.")] | None = None
    pump_load: (
        Annotated[
            ActivePower,
            Field(description="Load related to the usage of the pump."),
        ]
        | None
    ) = None

    @classmethod
    def example(cls) -> "HydroPumpedStorage":
        return HydroPumpedStorage(
            name="HydroPumpedStorage",
            active_power=ActivePower(100, "MW"),
            pump_load=ActivePower(100, "MW"),
            bus=ACBus.example(),
            prime_mover_type=PrimeMoversType.PS,
            storage_duration=Time(10, "h"),
            storage_capacity=UpDown(up=100, down=100),
            min_storage_capacity=Energy(10, "MWh"),
            pump_efficiency=Percentage(85, "%"),
            initial_volume=Energy(500, "MWh"),
            ext={"description": "Pumped hydro unit with 10 hour of duration"},
        )


class ThermalGen(Generator):
    """Class representing fuel based thermal generator."""

    fuel: Annotated[str, Field(description="Fuel category")] | None = None
    operation_cost: ThermalGenerationCost | None = None
    must_run: Annotated[bool, Field(description="If we need to force the dispatch of the device.")] = False


class ThermalStandard(ThermalGen):
    """Class representing a standard thermal generator."""

    status: bool = True
    ramp_limits: UpDown | None = None
    time_limits: UpDown | None = None
    fuel: ThermalFuels = ThermalFuels.OTHER

    @classmethod
    def example(cls) -> "ThermalStandard":
        return ThermalStandard(
            name="ThermalStandard",
            bus=ACBus.example(),
            fuel=ThermalFuels.NATURAL_GAS,
            active_power=100.0 * ureg.MW,
            ext={"Additional data": "Additional value"},
        )


class ThermalMultiStart(ThermalGen):
    """Class representing a MultiStart thermal generator."""

    # TODO(jwright2): Enhance #146 Update to include more comprehensive treatment of ThermalMultiStart
    # start_time_limits and start_types

    status: bool = True
    ramp_limits: UpDown | None = None
    time_limits: UpDown | None = None
    fuel: ThermalFuels = ThermalFuels.OTHER
    power_trajectory: StartShut | None = None
    start_time_limits: StartTimeLimits | None = None
    start_types: int | None = None

    @classmethod
    def example(cls) -> "ThermalMultiStart":
        return ThermalMultiStart(
            name="ThermalMultiStart",
            bus=ACBus.example(),
            fuel=ThermalFuels.COAL,
            active_power=100.0 * ureg.MW,
            power_trajectory=StartShut(startup=4.0, shutdown=4.0),
            start_time_limits=StartTimeLimits(hot=2.0, warm=6.0, cold=12.0),
            start_types=1,
            ext={"Additional data": "Additional value"},
        )


class Storage(Generator):
    """Default Storage class."""

    storage_duration: (
        Annotated[
            Time,
            Field(description="Storage duration in hours."),
        ]
        | None
    ) = None
    initial_energy: Annotated[Percentage, Field(description="Initial state of charge.")] | None = None
    min_storage_capacity: Annotated[Percentage, Field(description="Minimum state of charge")] = Percentage(
        0, "%"
    )
    max_storage_capacity: Annotated[Percentage, Field(description="Minimum state of charge")] = Percentage(
        100, "%"
    )
    storage_capacity: (
        Annotated[
            Energy,
            Field(description="Total water volume or percentage."),
        ]
        | None
    ) = None


class EnergyReservoirStorage(Storage):
    """Battery energy storage model."""

    storage_technology_type: StorageTechs
    operation_cost: StorageCost | None = None
    charge_efficiency: Annotated[Percentage, Field(ge=0, description="Charge efficiency.")] | None = None
    discharge_efficiency: Annotated[Percentage, Field(ge=0, description="Discharge efficiency.")] | None = (
        None
    )
    storage_level_limits: MinMax = MinMax(min=0, max=1)
    initial_storage_capacity_level: float
    input_active_power_limits: MinMax
    output_active_power_limits: MinMax
    efficiency: InputOutput


class HybridSystem(Device):
    """Representation of hybrid system with renewable generation, load, thermal generation and storage.

    This class is just a link between two components.
    For the implementation see:
    - https://github.com/NREL-Sienna/PowerSystems.jl/blob/main/src/models/HybridSystem.jl
    """

    storage_unit: Storage | None = None
    renewable_unit: RenewableGen | None = None
    thermal_unit: ThermalGen | None = None
    electric_load: PowerLoad | None = None



"""Helper functions for the exporters."""

from typing import Any
from collections.abc import Callable
from r2x.enums import ReserveType, ReserveDirection
import pint
from infrasys.base_quantity import BaseQuantity


def get_reserve_type(
    reserve_type: ReserveType, reserve_direction: ReserveDirection, reserve_types: dict[str, dict[str, str]]
) -> str:
    """Return the reserve type from a mapping.

    If not found, return default reserve type
    """
    for key, value in reserve_types.items():
        if value["type"] == reserve_type and value["direction"] == reserve_direction:
            return key
    return get_reserve_type(
        ReserveType[reserve_types["default"]["type"]],
        ReserveDirection[reserve_types["default"]["direction"]],
        reserve_types,
    )


def compose(
    *functions: Callable[[dict[str, Any]], dict[str, Any]],
) -> Callable[[dict[str, Any]], dict[str, Any]]:
    """
    Compose multiple functions, applying them sequentially to the input data.

    Parameters
    ----------
    functions : Callable[[Dict[str, Any]], Dict[str, Any]]
        Functions that transform a dictionary.

    Returns
    -------
    Callable[[Dict[str, Any]], Dict[str, Any]]
        A function that sequentially applies the provided functions to a dictionary.
    """

    def apply_all(data: dict[str, Any]) -> dict[str, Any]:
        for func in functions:
            data = func(data)
        return data

    return apply_all


def modify_components(
    *transform_functions: Callable[[dict[str, Any]], dict[str, Any]],
) -> Callable[[dict[str, Any]], dict[str, Any]]:
    """Apply multiple transformations to components."""
    return compose(*transform_functions)


def apply_property_map(component: dict[str, Any], property_map: dict[str, str]) -> dict[str, Any]:
    """Apply a key mapping to component keys.

    For each key in the `component` dictionary, if the key exists in the `property_map`,
    it will be replaced by the corresponding mapped key. Otherwise, the original key is used.

    Parameters
    ----------
    component : dict[str, Any]
        The original dictionary where keys represent component properties.
    property_map : dict[str, str]
        A dictionary mapping old property names to new property names.

    Returns
    -------
    dict[str, Any]
        A new dictionary where the keys have been remapped according to `property_map`.

    Examples
    --------
    >>> component = {"voltage": 230, "current": 10}
    >>> property_map = {"voltage": "v", "current": "i"}
    >>> apply_property_map(component, property_map)
    {'v': 230, 'i': 10}

    >>> component = {"voltage": 230, "resistance": 50}
    >>> property_map = {"voltage": "v", "current": "i"}
    >>> apply_property_map(component, property_map)
    {'v': 230, 'resistance': 50}
    """
    return {property_map.get(key, key): value for key, value in component.items()}


def apply_pint_deconstruction(component: dict[str, Any], unit_map: dict[str, str]) -> dict[str, Any]:
    """Get property magnitude from a pint Quantity.

    If unit_map is passed, convert to units specified

    Parameters
    ----------
    component: dict[str, Any]
        Dictionary representation of component. Typically created with `.model_dump()`
    unit_map: dict[str, str]
        Map to convert property to a desired units. Optional.
    """
    return {
        property: get_property_magnitude(property_value, to_unit=unit_map.get(property, None))
        for property, property_value in component.items()
    }


def apply_valid_properties(
    component: dict[str, Any], valid_properties: list[str], add_name: bool = False
) -> dict[str, Any]:
    """Filter a component dictionary to only include keys that are in the valid properties list.

    Parameters
    ----------
    component : dict of str to Any
        A dictionary representing the component with properties as keys.
    valid_properties : list of str
        A list of valid property names. Only keys present in this list will be kept in the output.

    Returns
    -------
    dict
        A new dictionary with only the properties from `valid_properties`.

    Examples
    --------
    >>> component = {"voltage": 230, "current": 10, "resistance": 50}
    >>> valid_properties = ["voltage", "current"]
    >>> apply_valid_properties(component, valid_properties)
    {'voltage': 230, 'current': 10}
    """
    if add_name:
        default_properties = ["name"]
        return {
            key: value
            for key, value in component.items()
            if key in valid_properties or key in default_properties
        }
    return {key: value for key, value in component.items() if key in valid_properties}


def apply_unnest_key(component: dict[str, Any], key_map: dict[str, Any]) -> dict[str, Any]:
    """Unnest specific nested dictionary values based on a provided key map.

    This function processes a dictionary, potentially containing nested dictionaries,
    and unnests specific values based on a provided key map. For each key in the input
    dictionary that has a corresponding entry in the key map, if the value is a dictionary,
    it extracts the value using the mapped key.

    Parameters
    ----------
    component : dict[str, Any]
        The input dictionary to process.
    key_map : dict[str, Any]
        A dictionary mapping keys in the input dictionary to keys in nested dictionaries.

    Returns
    -------
    dict[str, Any]
        A new dictionary with unnested values based on the key map.

    Examples
    --------
    >>> component = {
    ...     "name": "Example",
    ...     "config": {"type": "A", "value": 10},
    ...     "data": {"content": "Some data"},
    ... }
    >>> key_map = {"config": "type", "data": "content"}
    >>> apply_unnest_key(component, key_map)
    {'name': 'Example', 'config': 'A', 'data': 'Some data'}

    Notes
    -----
    - If a key in the input dictionary is not in the key map, its value remains unchanged.
    - If a key is in the key map but the corresponding value in the input dictionary
      is not a dictionary, the value remains unchanged.
    - If a key is in the key map and the corresponding value is a dictionary, but the
      mapped key is not in this nested dictionary, the result for this key will be None.
    """
    if not key_map:
        return component
    return {
        key: value if not isinstance(value, dict) else value.get(key_map.get(key), value)
        for key, value in component.items()
    }


def apply_default_value(component: dict[str, Any], default_value_map: dict[str, Any]) -> dict[str, Any]:
    """Unnest specific nested dictionary values based on a provided key map.

    This function iterates over a dictionary (`component`) and applies default values
    to its missing or `None` values using a `default_value_map`. For each key in
    `default_value_map`, if the key is not present in `component` or has a `None` value,
    it will be assigned the corresponding default value from `default_value_map`.

    Parameters
    ----------
    component : dict[str, Any]
        The input dictionary to process.
    default_value_map : dict[str, Any]
        A dictionary specifying default values. Each key in this dictionary corresponds
        to a key in `component`, and the associated value is the default value to apply
        if the key is missing or `None` in `component`.

    Returns
    -------
    dict[str, Any]
        A new dictionary with unnested values based on the key map.

    Examples
    --------
    >>> component = {"name": "example", "year": None}
    >>> default_value_map = {"year": 2024, "month": "M10"}
    >>> apply_default_value(component, default_value_map)
    {'name': 'example', 'year': 2024, 'month': 'M10'}

    >>> component = {"name": "example", "year": 2023}
    >>> default_value_map = {"year": 2024, "month": "M10"}
    >>> apply_default_value(component, default_value_map)
    {'name': 'example', 'year': 2023, 'month': 'M10'}
    """
    if not default_value_map:
        return component
    for key in default_value_map:
        if key not in component:
            component[key] = default_value_map[key]
        if component.get(key, None) is None:
            component[key] = default_value_map[key]
    return component


def get_property_magnitude(property_value, to_unit: str | None = None) -> Any:
    """Return magnitude with the given units for a pint Quantity.

    Parameters
    ----------
    property_name

    property_value
        pint.Quantity to extract magnitude from
    to_unit
        String that contains the unit conversion desired. Unit must be compatible.

    Returns
    -------
    float
        Magnitude representation of the `pint.Quantity` or original value.
    """
    if not isinstance(property_value, pint.Quantity | BaseQuantity):
        return property_value
    if to_unit:
        unit = to_unit.replace("$", "usd")  # Dollars are named usd on pint
        property_value = property_value.to(unit)
    return property_value.magnitude


def apply_flatten_key(d: dict[str, Any], keys_to_flatten: set[str]) -> dict[str, Any]:
    """Flatten the specified keys in a dictionary by merging their sub-keys into the main dictionary
    with the key's name prefixed to each sub-key.

    Parameters
    ----------
    d : dict
        The input dictionary, where some values are dictionaries to be flattened.
    keys_to_flatten : list of str
        The keys in the dictionary `d` to be flattened if they contain sub-keys.

    Returns
    -------
    dict
        A new dictionary with the selected keys flattened. Other keys remain unchanged.

    Examples
    --------
    >>> d = {"x": {"min": 1, "max": 2}, "y": {"min": 5, "max": 10}, "z": 42}
    >>> flatten_selected_keys(d, ["x"])
    {'x_min': 1, 'x_max': 2, 'y': {'min': 5, 'max': 10}, 'z': 42}

    >>> flatten_selected_keys(d, ["y"])
    {'x': {'min': 1, 'max': 2}, 'y_min': 5, 'y_max': 10, 'z': 42}
    """
    flattened_dict = {}

    for key, value in d.items():
        if key in keys_to_flatten and isinstance(value, dict):
            for inner_key, inner_value in value.items():
                flattened_dict[f"{key}_{inner_key}"] = inner_value
        else:
            flattened_dict[key] = value

    return flattened_dict


def apply_extract_key(d: dict[str, Any], key: str, keys_to_extract: set[str]) -> dict[str, Any]:
    """Extract keys from a nested dictionary and put them in the first level if specific conditions are met.

    Parameters
    ----------
    d : dict
        The input dictionary that may contain nested dictionaries.
    key : str
        The key in the input dictionary whose value should be a nested dictionary.
    keys_to_extract : set[str]
        The set of keys to extract from the nested dictionary.

    Returns
    -------
    dict
        If conditions are met: a new dictionary containing all original key-value pairs
        plus the extracted key-value pairs at the top level.
        If conditions are not met: returns the input dictionary unchanged.

    Notes
    -----
    The function will return the input dictionary unchanged if any of these conditions are met:
    - The specified key is not in the input dictionary
    - Any of keys_to_extract already exist in the top level of input dictionary
    - None of keys_to_extract exist in the nested dictionary

    Examples
    --------
    >>> component = {"name": "Gen01", "rating": {"up": 100, "down": -100}}
    >>> apply_extract_key(component, "rating", {"up"})
    {'name': 'Gen01', 'rating': {'up': 100, 'down': -100}, 'up': 100}

    >>> # Returns unchanged when extracted key already exists at top level
    >>> d = {"name": "Gen01", "rating": {"up": 100}, "up": 50}
    >>> apply_extract_key(d, "rating", {"up"})
    {'name': 'Gen01', 'rating': {'up': 100}, 'up': 50}

    >>> # Returns unchanged when no keys_to_extract exist in nested dict
    >>> apply_extract_key(component, "rating", {"middle"})
    {'name': 'Gen01', 'rating': {'up': 100, 'down': -100}}
    """
    if key not in d.keys() or any(k in d.keys() for k in keys_to_extract):
        return d

    if not any(k in d[key].keys() for k in keys_to_extract):
        return d

    extracted_keys = {key: value for key, value in d[key].items() if key in keys_to_extract}
    return {**d, **extracted_keys}



"""Concrete class to handler parser models.

This module provides the abstract class to create parser objects.
"""

# System packages
import inspect
import json
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, TypeVar

import polars as pl

# Third-party packages
from infrasys.component import Component
from loguru import logger
from plexosdb import XMLHandler
from pydantic import ValidationError

# Local packages
from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.exceptions import R2XParserError
from r2x.utils import check_file_exists

from .handler_utils import csv_handler, h5_handler
from .polars_helpers import pl_filter_by_weather_year, pl_filter_by_year, pl_rename

FILE_PARSING_KWARGS = {
    "absolute_fpath",
    "keep_case",
    "use_filter_functions",
    "column_mapping",
    "solve_year",
    "weather_year",
    "filter_by_weather_year",
}


@dataclass
class BaseParser(ABC):
    """Class that defines the shared methods of parsers.

    Note
    ----
    This class is meant to be use for developing new parsers. Do not use it directly.

    Attributes
    ----------
    config: Scenario
        Scenario configuration
    data: dict
        We save each file read in a data dictionary

    Methods
    -------
    get_data(key='load')
        Return the parsed data for load.
    read_file(fpath="load.csv")
        Read load data.
    parse_data
        Read all files from a configuration file map.
    """

    config: Scenario
    data: dict = field(default_factory=dict)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(Files parsed: {len(self.data)})"

    def get_data(self, key: str) -> Any:
        """Return data."""
        if key not in self.data:
            raise KeyError(f"Key `{key}` not found in data dictionary.")
        return self.data[key]

    def read_file(
        self,
        fpath: Path | str,
        filter_funcs: list[Callable] | None = None,
        use_filter_functions: bool = True,
        **kwargs,
    ):
        """Read input model data from the file system.

        Currently supported formats:
            - .csv
            - .h5
            - .xml
        More to come!

        Parameters
        ----------
        fpath: Path, str
            Absolute location of the file in the system
        filter_func: List
            Filter functions to apply

        """
        data = file_handler(fpath, parser_class=type(self).__name__, **kwargs)
        if data is None:
            return

        if kwargs.get("filter_by_weather_year") and isinstance(filter_funcs, list):
            filter_funcs.append(pl_filter_by_weather_year)

        if use_filter_functions and isinstance(filter_funcs, list):
            for func in filter_funcs:
                data = func(data, **kwargs)
        return data

    def parse_data(
        self,
        *,
        folder: Path,
        fmap: dict,
        filter_funcs: list[Callable] | None = None,
        **kwargs,
    ) -> bool:
        """Parse all the data for the given translation."""
        logger.trace("Parsing data for {}", self.__class__.__name__)
        _fmap = deepcopy(fmap)

        files_to_parse = {key: value for key, value in _fmap.items() if value is not None}
        if not len(files_to_parse) > 0:
            msg = "Not a single valid entry found on the fmap configuration."
            raise R2XParserError(msg)

        for dname, data in files_to_parse.items():
            fname = data["fname"]
            is_optional = data.get("optional")
            data = {**data, **kwargs}
            file_parsing_kwargs = {key: value for key, value in data.items() if key in FILE_PARSING_KWARGS}

            # If we pass an absolute path we check that it exists.
            if fpath := data.get("absolute_fpath"):
                fpath = Path(fpath)

            if not fpath:
                fpath = check_file_exists(fname=fname, run_folder=folder, folder=data.get("folder"))

            if not fpath and not is_optional:
                msg = (
                    f"Mandatory file {data['fname']} not found at {fpath}. "
                    "Check that the provided path is correct."
                )
                raise R2XParserError(msg)

            if fpath is not None:
                logger.debug("Loading file {} from {}", dname, fpath)
                self.data[dname] = self.read_file(fpath, filter_funcs=filter_funcs, **file_parsing_kwargs)
        return True

    @abstractmethod
    def build_system(self) -> System:
        """Create the infra_sys model."""


class PCMParser(BaseParser):
    """Class defining shared methods for PCM (currently plexos and sienna) parsers."""

    pass


def file_handler(
    fpath: Path | str, parser_class: str | None = None, optional: bool = False, **kwargs
) -> pl.LazyFrame | pl.DataFrame | Sequence | XMLHandler | None:
    """Return FileHandler based on file extension.

    Raises
    ------
    FileNotFoundError
        If the file is not found.
    NotImplementedError
        If the file format is not yet supported or the file format is not supported for the parser.
    """
    logger.trace("Attempting to read: {}", fpath)
    if not isinstance(fpath, Path):
        fpath = Path(fpath)

    logger.trace("Reading {}", fpath)
    match fpath.suffix:
        case ".csv":
            return csv_handler(fpath, **kwargs)
        case ".h5":
            assert parser_class is not None
            return h5_handler(fpath, parser_class=parser_class, **kwargs)
        case ".xml":
            class_kwargs = {
                key: value for key, value in kwargs.items() if key in inspect.signature(XMLHandler).parameters
            }
            return XMLHandler.parse(fpath=fpath, **class_kwargs)
        case ".json":
            with open(fpath) as json_file:
                data = json.load(json_file)
            return data
        case _:
            raise NotImplementedError(f"File {fpath.suffix = } not yet supported.")


ParserClass = TypeVar("ParserClass", bound=BaseParser)


def get_parser_data(
    config: Scenario,
    parser_class: Callable,
    filter_funcs: list[Callable] | None = None,
    **kwargs,
) -> BaseParser:
    """Return parsed system.

    This function will create the ReEDS DataPortal and populate with the most
    common set of data needed


    Paremters
    ---------
    config Scenario configuration class
    parser
        Parser to process
    filter_func
        Functions that will applied to read_data process

    Other Parameters
    ----------------
    kwargs
        year
            For filtering by solve year
        year_column
            To change the column to apply the filter
        column_mapping
            For renaming columns

    See Also
    --------
    BaseParser
    pl_filter_year
    pl_lower_case
    pl_rename
    """
    logger.debug("Creating {} instance.", parser_class.__name__)

    parser = parser_class(config=config, **kwargs)

    # Functions relative to the parser.
    # NOTE: At some point we are going to migrate this out, but this sound like a good standard set
    if (filter_funcs is None) and (parser_class.__name__ == "ReEDSParser"):
        logger.trace("Using default filter functions")
        filter_funcs = [pl_rename, pl_filter_by_year]

    # Adding special case for Plexos parser
    if model := getattr(config, "model", False):
        kwargs["model"] = model

    # Parser data
    assert config.input_config
    parser.parse_data(
        folder=config.run_folder,
        filter_funcs=filter_funcs,
        **{**config.input_config.__dict__, **kwargs},
    )

    # Create system
    logger.debug("Starting creation of system: {}", config.name)

    return parser


def create_model_instance(
    model_class: type["Component"], skip_validation: bool = False, **field_values
) -> Any:
    """Create R2X model instance."""
    valid_fields = {
        key: value
        for key, value in field_values.items()
        if key in model_class.model_fields
        if value is not None
    }
    if skip_validation:
        try:
            return model_class.model_validate(valid_fields)
        except ValidationError:
            return model_class.model_construct(**valid_fields)
    return model_class.model_validate(valid_fields)



"""Create PLEXOS model from translated ReEDS data."""

import string
import uuid
from argparse import ArgumentParser
from collections.abc import Callable
from functools import partial
from importlib.resources import files
from typing import Any

from infrasys.component import Component
from loguru import logger
from plexosdb import PlexosSQLite
from plexosdb.enums import ClassEnum, CollectionEnum

from r2x.config_models import PlexosConfig, ReEDSConfig
from r2x.enums import ReserveType
from r2x.exporter.handler import BaseExporter, get_export_properties, get_export_records
from r2x.exporter.utils import (
    apply_extract_key,
    apply_flatten_key,
    apply_pint_deconstruction,
    apply_property_map,
    apply_valid_properties,
    get_reserve_type,
)
from r2x.models import (
    ACBus,
    Emission,
    EnergyReservoirStorage,
    Generator,
    HydroDispatch,
    HydroEnergyReservoir,
    HydroPumpedStorage,
    InterruptiblePowerLoad,
    LoadZone,
    MonitoredLine,
    PowerLoad,
    RenewableDispatch,
    RenewableNonDispatch,
    Reserve,
    ThermalStandard,
    Transformer2W,
    TransmissionInterface,
)
from r2x.models.branch import Line
from r2x.models.utils import Constraint
from r2x.units import get_magnitude
from r2x.utils import custom_attrgetter, get_enum_from_string, read_json

NESTED_ATTRIBUTES = {"ext", "bus", "services"}
TIME_SERIES_PROPERTIES = ["Min Provision", "Static Risk"]
DEFAULT_XML_TEMPLATE = "master_9.2R6_btu.xml"
EXT_PROPERTIES = {"UoS Charge", "Fixed Load"}
KEYS_TO_FLATTEN = {"active_power_limits", "active_power_flow_limits", "storage_capacity", "flow_limits"}


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--master-file",
        required=False,
        help="Plexos master file to use as template.",
    )


class PlexosExporter(BaseExporter):
    """Plexos exporter class."""

    def __init__(
        self,
        *args,
        plexos_scenario: str = "default",
        database_manager=None,
        xml_fname: str | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        assert self.config.output_config
        assert self.config.input_config

        if not isinstance(self.config.output_config, PlexosConfig):
            msg = (
                f"Output config is of type {type(self.config.output_config)}. "
                "It should be type of `PlexosConfig`."
            )
            raise TypeError(msg)

        # Do not allow multiple years for the solve year.
        if isinstance(self.config.input_config, ReEDSConfig) and isinstance(
            self.config.input_config.solve_year, list
        ):
            msg = "Multiple solve years are not supported yet."
            raise NotImplementedError(msg)

        # Map relevant input configuration to output configuration
        self.output_config = self.config.input_config.to_class(PlexosConfig, self.config.output_config)
        self.plexos_scenario = plexos_scenario or self.output_config.model_name
        if not xml_fname and not (xml_fname := getattr(self.output_config, "master_file", None)):
            xml_fname = files("r2x.defaults").joinpath(DEFAULT_XML_TEMPLATE)  # type: ignore
            logger.debug("Using default XML template.")

        # Initialize PlexosDB
        self._db_mgr = database_manager or PlexosSQLite(xml_fname=xml_fname)
        self.plexos_scenario_id = self._db_mgr.get_scenario_id(scenario_name=plexos_scenario)

        self._setup_plexos_configuration()

    def _setup_plexos_configuration(self) -> None:
        self.property_map = self.output_config.defaults["plexos_property_map"]
        self.valid_properties = self.output_config.defaults["valid_properties"]
        self.default_units = self.output_config.defaults["default_units"]
        self.reserve_types = self.output_config.defaults["reserve_types"]

        self.simulation_objects = self.output_config.defaults["simulation_objects"]
        self.static_horizon_type = self.output_config.defaults["static_horizon_type"]
        self.static_horizons = self.output_config.defaults[self.static_horizon_type]
        self.static_model_type = self.output_config.defaults["static_model_type"]
        self.static_models = self.output_config.defaults[self.static_model_type]
        self.plexos_reports_fpath = self.output_config.defaults["plexos_reports"]

        # Set modeling years that will be used.
        assert isinstance(self.output_config.model_year, int)
        self.model_year: int = self.output_config.model_year
        self.weather_year: int = self.output_config.horizon_year or self.model_year

    def run(self, *args, new_database: bool = True, **kwargs) -> "PlexosExporter":
        """Run the exporter."""
        logger.info("Starting {}", self.__class__.__name__)

        self.time_series_to_csv(config=self.config, system=self.system, reference_year=self.weather_year)

        # If starting w/o a reference file we add our custom models and objects
        if new_database:
            self._add_simulation_objects()
            self._add_horizons()
            self._add_models()
            self._add_reports()
        self.add_constraints()
        self.add_topology()
        self.add_lines()
        self.add_transformers()
        self.add_interfaces()
        self.add_emissions()
        self.add_reserves()
        self.add_batteries()
        self.add_generators()
        self.add_storage()

        self._db_mgr.to_xml(fpath=f"{self.output_folder}/{self.config.name}.xml")

        return self

    def _get_time_series_properties(self, component):  # noqa: C901
        """Add time series object to certain plexos properties."""
        if not self.system.has_time_series(component):
            return

        if len(self.system.list_time_series(component)) > 1:
            # NOTE:@pedro this is a temporary fix for the multiple time series issue.
            return

        ts_metadata = self.system.get_time_series(component)

        config_dict = self.config.__dict__
        config_dict["name"] = self.config.name
        csv_fname = config_dict.get("time_series_fname", "${component_type}_${name}_${weather_year}.csv")
        string_template = string.Template(csv_fname)
        config_dict["component_type"] = f"{component.__class__.__name__}_{ts_metadata.variable_name}"
        config_dict["weather_year"] = self.weather_year
        csv_fname = string_template.safe_substitute(config_dict)
        csv_fpath = self.ts_directory / csv_fname
        time_series_property: dict[str, Any] = {"Data File": str(csv_fpath)}

        # Property with time series can change. Validate this option based on the component type.
        match component:
            case InterruptiblePowerLoad():
                time_series_property["Fixed Load"] = "0"
            case PowerLoad():
                time_series_property["Load"] = "0"
            case RenewableDispatch():
                time_series_property["Rating"] = "0"
                time_series_property["Load Subtracter"] = "0"
            case RenewableNonDispatch():
                time_series_property["Rating"] = "0"
                time_series_property["Load Subtracter"] = "0"
            case Reserve():
                match component.reserve_type:
                    case ReserveType.SPINNING:
                        time_series_property["Min Provision"] = "0"
                    case ReserveType.FLEXIBILITY:
                        time_series_property["Min Provision"] = "0"
                    case ReserveType.REGULATION:
                        time_series_property["Min Provision"] = "0"
                    case _:
                        raise NotImplementedError(f"Reserve {component.type} not supported")
            case HydroDispatch():
                variable_name = self.system.get_time_series(component).variable_name
                if not variable_name:
                    return None
                property_name = self.property_map.get(variable_name, None)
                if property_name:
                    time_series_property[property_name] = "0"
            case HydroEnergyReservoir():
                time_series_property["Fixed Load"] = "0"
            case ThermalStandard():
                variable_name = self.system.get_time_series(component).variable_name
                if not variable_name:
                    return None
                property_name = self.property_map.get(variable_name, None)
                if property_name:
                    time_series_property[property_name] = "0"
            case _:
                raise NotImplementedError(f"Time Series for {component.label} not supported yet.")
        return time_series_property

    def insert_component_properties(
        self,
        component_type: type["Component"],
        /,
        *,
        parent_class: ClassEnum,
        parent_object_name: str = "System",
        collection: CollectionEnum,
        child_class: ClassEnum | None = None,
        filter_func: Callable | None = None,
        scenario: str | None = None,
        records: list[dict] | None = None,
        exclude_fields: set[str] | None = NESTED_ATTRIBUTES,
    ) -> None:
        """Bulk insert properties from selected component type."""
        logger.debug("Adding {} table properties...", component_type.__name__)
        scenario = scenario or self.plexos_scenario
        if not records:
            records = [
                component.model_dump(
                    exclude_none=True, exclude=exclude_fields, mode="python", serialize_as_any=True
                )
                for component in self.system.get_components(component_type, filter_func=filter_func)
            ]

        if not records:
            logger.warning("No components found for type {}", component_type)
            return

        collection_properties = self._db_mgr.get_valid_properties(
            collection, parent_class=parent_class, child_class=child_class
        )
        # property_names = [key[0] for key in collection_properties]
        match component_type.__name__:
            case "EnergyReservoirStorage":
                custom_map = {"active_power": "Max Power", "storage_capacity": "Capacity"}
            case "Line":
                custom_map = {"rating": "Max Flow"}
            case _:
                custom_map = {}
        property_map = self.property_map | custom_map

        export_records = get_export_records(
            records,
            partial(apply_operation_cost),
            partial(apply_extract_key, key="ext", keys_to_extract=EXT_PROPERTIES),
            partial(
                apply_flatten_key,
                keys_to_flatten=KEYS_TO_FLATTEN,
            ),
            partial(apply_property_map, property_map=property_map),
            partial(apply_pint_deconstruction, unit_map=self.default_units),
            partial(apply_valid_properties, valid_properties=collection_properties, add_name=True),
        )
        self._db_mgr.add_property_from_records(
            export_records,
            parent_class=parent_class,
            parent_object_name=parent_object_name,
            collection=collection,
            scenario=scenario,
        )

    def add_component_category(
        self,
        component_type: type["Component"],
        class_enum: ClassEnum,
        category_attribute: str = "category",
        filter_func: Callable | None = None,
    ):
        """Add all categories for a component type."""
        component_categories = {
            custom_attrgetter(component, category_attribute)
            for component in self.system.get_components(component_type, filter_func=filter_func)
        }

        existing_rank = self._db_mgr.get_category_max_id(class_enum)
        class_id = self._db_mgr.get_class_id(class_enum)
        categories = [
            (class_id, rank, category or "")
            for rank, category in enumerate(
                sorted(component_categories, key=lambda x: (x is None, x)), start=existing_rank + 1
            )
        ]

        # Maybe replace `t_category` with right schema.
        with self._db_mgr._conn as conn:
            conn.executemany("INSERT into t_category(class_id, rank, name) values (?,?,?)", categories)
        return

    def bulk_insert_objects(
        self,
        component_type: type["Component"],
        class_enum: ClassEnum,
        collection_enum: CollectionEnum,
        category_attribute: str = "category",
        name_prefix: str = "",
        filter_func: Callable | None = None,
    ):
        """Bulk insert objects to the database."""
        logger.debug("Adding plexos objects for component type {}", component_type.__name__)

        query = """
        SELECT
            t_category.name
            ,t_category.category_id
        FROM
            t_category
        LEFT JOIN
            t_class ON t_class.class_id = t_category.class_id
        WHERE
            t_class.name = :class_name
        """
        categories_ids = {
            key: value for key, value in self._db_mgr.query(query, params={"class_name": class_enum})
        }
        class_id = self._db_mgr.get_class_id(class_enum)

        objects = [
            (
                class_id,
                component.name + name_prefix,
                self._get_category_id(component, category_attribute, categories_ids),
                str(uuid.uuid4()),
            )
            for component in self.system.get_components(component_type, filter_func=filter_func)
        ]
        object_names = tuple(d[1] for d in objects)
        if not objects:
            logger.warning("No components found for type' {}", component_type)
            return
        with self._db_mgr._conn as conn:
            conn.executemany(
                "INSERT into t_object(class_id, name, category_id, GUID) values (?,?,?,?)", objects
            )

        # Add system membership
        system_object_id = self._db_mgr.get_object_id("System", class_name=ClassEnum.System)
        system_class_id = self._db_mgr.get_class_id(ClassEnum.System)
        collection_id = self._db_mgr.get_collection_id(
            collection_enum, parent_class=ClassEnum.System, child_class=class_enum
        )
        objects_placeholders = ", ".join("?" * len(object_names))
        membership_query = f"""
            INSERT into t_membership(
              parent_class_id, parent_object_id,
              collection_id, child_class_id, child_object_id
            )
            SELECT
                ? as parent_class_id,
                ? as parent_object_id,
                ? as collection_id,
                ? as child_class_id,
                object_id as child_object_id
            FROM
              t_object
            WHERE
              class_id = ?
              and t_object.name in ({objects_placeholders})
        """
        query_parameters = (
            system_class_id,
            system_object_id,
            collection_id,
            class_id,
            class_id,
            *object_names,
        )
        self._db_mgr.execute_query(membership_query, query_parameters)

        # Enable all classes that we add into the database.
        # Some are disabled by default.
        self._db_mgr.execute_query(f"UPDATE t_class SET is_enabled=1 WHERE t_class.name='{class_enum}'")
        return

    def add_topology(self) -> None:
        """Create network topology on Plexos."""
        # Adding Regions
        self.add_component_category(ACBus, category_attribute="area.name", class_enum=ClassEnum.Region)
        self.bulk_insert_objects(
            ACBus,
            category_attribute="area.name",
            class_enum=ClassEnum.Region,
            collection_enum=CollectionEnum.Regions,
        )
        self.insert_component_properties(
            ACBus, parent_class=ClassEnum.System, collection=CollectionEnum.Regions
        )
        for bus in self.system.get_components(ACBus, filter_func=lambda x: x.ext):
            collection_properties = self._db_mgr.get_valid_properties(
                collection=CollectionEnum.Zones, parent_class=ClassEnum.System, child_class=ClassEnum.Zone
            )
            properties = get_export_properties(
                bus.ext,
                partial(apply_property_map, property_map=self.property_map),
                partial(apply_pint_deconstruction, unit_map=self.default_units),
                partial(apply_valid_properties, valid_properties=collection_properties),
            )
            if properties:
                for property_name, property_value in properties.items():
                    self._db_mgr.add_property(
                        bus.name,
                        property_name,
                        property_value,
                        object_class=ClassEnum.Region,
                        collection=CollectionEnum.Regions,
                        scenario=self.plexos_scenario,
                    )

        # Adding Zones
        # self.add_component_category(LoadZone, class_enum=ClassEnum.Zone)
        self.bulk_insert_objects(LoadZone, class_enum=ClassEnum.Zone, collection_enum=CollectionEnum.Zones)
        self.insert_component_properties(
            LoadZone, parent_class=ClassEnum.System, collection=CollectionEnum.Zones
        )

        # Adding nodes
        # NOTE: For nodes, we do not add category.
        # self.add_component_category(ACBus, class_enum=ClassEnum.Node)
        self.bulk_insert_objects(ACBus, class_enum=ClassEnum.Node, collection_enum=CollectionEnum.Nodes)
        self.insert_component_properties(
            ACBus, parent_class=ClassEnum.System, collection=CollectionEnum.Nodes
        )

        # Add node memberships to zone and regions.
        # On our default Plexos translation, both Zones and Regions are child of the Node class.
        for bus in self.system.get_components(ACBus):
            bus_load_zone = bus.load_zone
            self._db_mgr.add_membership(
                bus.name,
                bus.name,  # Zone has the same name
                parent_class=ClassEnum.Node,
                child_class=ClassEnum.Region,
                collection=CollectionEnum.Region,
            )
            if bus_load_zone is None:
                continue
            self._db_mgr.add_membership(
                bus.name,
                bus_load_zone.name,
                parent_class=ClassEnum.Node,
                child_class=ClassEnum.Zone,
                collection=CollectionEnum.Zone,
            )

        # Adding load time series
        logger.debug("Adding load time series properties")
        for component in self.system.get_components(
            PowerLoad, filter_func=lambda component: self.system.has_time_series(component)
        ):
            time_series_properties = self._get_time_series_properties(component)
            if time_series_properties:
                text = time_series_properties.pop("Data File")
                for property_name, property_value in time_series_properties.items():
                    self._db_mgr.add_property(
                        component.bus.name,
                        property_name,
                        property_value,
                        object_class=ClassEnum.Region,
                        collection=CollectionEnum.Regions,
                        scenario=self.plexos_scenario,
                        text={"Data File": text},
                    )
        return

    def add_lines(self) -> None:
        """Add transmission lines that connect topology elements."""
        # Adding Lines
        # NOTE: The default line on Plexos is a `MonitoredLine` without category. If we need to add a category
        # in the future, we will uncomment the line below with the pertinent category name.
        self.add_component_category(MonitoredLine, class_enum=ClassEnum.Line)
        self.bulk_insert_objects(Line, class_enum=ClassEnum.Line, collection_enum=CollectionEnum.Lines)
        self.insert_component_properties(Line, parent_class=ClassEnum.System, collection=CollectionEnum.Lines)
        self.bulk_insert_objects(
            MonitoredLine, class_enum=ClassEnum.Line, collection_enum=CollectionEnum.Lines
        )
        self.insert_component_properties(
            MonitoredLine, parent_class=ClassEnum.System, collection=CollectionEnum.Lines
        )

        # Add additional properties if any and membersips
        collection_properties = self._db_mgr.get_valid_properties(
            collection=CollectionEnum.Lines, parent_class=ClassEnum.System, child_class=ClassEnum.Line
        )
        for line in self.system.get_components(MonitoredLine, Line):
            properties = get_export_properties(
                line.ext,
                partial(apply_property_map, property_map=self.property_map),
                partial(apply_pint_deconstruction, unit_map=self.default_units),
                partial(apply_valid_properties, valid_properties=collection_properties),
            )
            for property_name, property_value in properties.items():
                self._db_mgr.add_property(
                    line.name,
                    property_name,
                    property_value,
                    object_class=ClassEnum.Line,
                    collection=CollectionEnum.Lines,
                    scenario=self.plexos_scenario,
                )
            self._db_mgr.add_membership(
                line.name,
                line.from_bus.name,
                parent_class=ClassEnum.Line,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.NodeFrom,
            )
            self._db_mgr.add_membership(
                line.name,
                line.to_bus.name,
                parent_class=ClassEnum.Line,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.NodeTo,
            )
        return

    def add_transformers(self) -> None:
        """Add Transformer objects to the database."""
        self.add_component_category(Transformer2W, class_enum=ClassEnum.Transformer)
        self.bulk_insert_objects(
            Transformer2W, class_enum=ClassEnum.Transformer, collection_enum=CollectionEnum.Transformers
        )
        self.insert_component_properties(
            Transformer2W, parent_class=ClassEnum.System, collection=CollectionEnum.Transformers
        )
        for transformer in self.system.get_components(Transformer2W):
            self._db_mgr.add_membership(
                transformer.name,
                transformer.from_bus.name,
                parent_class=ClassEnum.Transformer,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.NodeFrom,
            )
            self._db_mgr.add_membership(
                transformer.name,
                transformer.to_bus.name,
                parent_class=ClassEnum.Transformer,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.NodeTo,
            )
        return

    def add_interfaces(self) -> None:
        """Add transmission interfaces."""
        self.bulk_insert_objects(
            TransmissionInterface,
            class_enum=ClassEnum.Interface,
            collection_enum=CollectionEnum.Interfaces,
        )
        self.insert_component_properties(
            TransmissionInterface, parent_class=ClassEnum.System, collection=CollectionEnum.Interfaces
        )

    def add_constraints(self) -> None:
        """Add custom constraints."""
        self.bulk_insert_objects(
            Constraint,
            class_enum=ClassEnum.Constraint,
            collection_enum=CollectionEnum.Constraints,
        )
        collection_properties = self._db_mgr.get_valid_properties(
            collection=CollectionEnum.Constraints,
            parent_class=ClassEnum.System,
            child_class=ClassEnum.Constraint,
        )
        for constraint in self.system.get_components(Constraint):
            properties = get_export_properties(
                constraint.ext,
                partial(apply_property_map, property_map=self.property_map),
                partial(apply_pint_deconstruction, unit_map=self.default_units),
                partial(apply_valid_properties, valid_properties=collection_properties),
            )

            if properties:
                for property_name, property_value in properties.items():
                    self._db_mgr.add_property(
                        constraint.name,
                        property_name,
                        property_value,
                        object_class=ClassEnum.Constraint,
                        collection=CollectionEnum.Constraints,
                        scenario=self.plexos_scenario,
                    )

        return

    def add_emissions(self) -> None:
        """Add emission objects to the database."""
        logger.debug("Adding Emission objects...")
        self._db_mgr.execute_query(
            f"UPDATE t_class SET is_enabled=1 WHERE t_class.name='{ClassEnum.Emission}'"
        )
        # Getting all unique emission types (e.g., CO2, NOX) from the emissions objects.
        # NOTE: On Plexos, we need to add each emission type individually to the Emission class
        emission_types = set(
            map(lambda x: x.emission_type, list(self.system.get_supplemental_attributes(Emission)))
        )
        for emission_type in emission_types:
            self._db_mgr.add_object(
                emission_type,
                ClassEnum.Emission,
                CollectionEnum.Emissions,
            )

            # Add emission caps from emission_cap.py if added.
            emission_constraint_name = f"Annual_{emission_type}_cap"
            collection_properties = self._db_mgr.get_valid_properties(
                collection=CollectionEnum.Constraints,
                parent_class=ClassEnum.Emission,
                child_class=ClassEnum.Constraint,
            )
            for constraint in self.system.get_components(
                Constraint, filter_func=lambda x: x.name == emission_constraint_name
            ):
                self._db_mgr.add_membership(
                    emission_type,
                    constraint.name,
                    parent_class=ClassEnum.Emission,
                    child_class=ClassEnum.Constraint,
                    collection=CollectionEnum.Constraints,
                )
                properties = get_export_properties(
                    constraint.ext[emission_type],
                    partial(apply_property_map, property_map=self.property_map),
                    partial(apply_pint_deconstruction, unit_map=self.default_units),
                    partial(apply_valid_properties, valid_properties=collection_properties),
                )
                if properties:
                    for property_name, property_value in properties.items():
                        self._db_mgr.add_property(
                            constraint.name,
                            property_name,
                            property_value,
                            object_class=ClassEnum.Constraint,
                            parent_object_name=emission_type,
                            parent_class=ClassEnum.Emission,
                            collection=CollectionEnum.Constraints,
                            scenario=self.plexos_scenario,
                        )
        return

    def add_reserves(self) -> None:
        """Add system reserves to the database.

        This method only appends the information for the type of reserves. Each
        generator that contributes to the reserve will add the membership for
        the type of reserve as a child membership.
        """
        # Adding reserves
        # self.add_component_category(Reserve, class_enum=ClassEnum.Reserve)
        self.bulk_insert_objects(
            Reserve, class_enum=ClassEnum.Reserve, collection_enum=CollectionEnum.Reserves
        )

        self.insert_component_properties(
            Reserve,
            parent_class=ClassEnum.System,
            collection=CollectionEnum.Reserves,
            exclude_fields=NESTED_ATTRIBUTES | {"max_requirement"},
        )
        for reserve in self.system.get_components(Reserve):
            properties: dict[str, Any] = {}
            properties["Type"] = get_reserve_type(
                reserve_type=reserve.reserve_type,
                reserve_direction=reserve.direction,
                reserve_types=self.reserve_types,
            )
            properties["Is Enabled"] = "-1" if reserve.available else "0"
            properties["Mutually Exclusive"] = True

            for property, value in properties.items():
                self._db_mgr.add_property(
                    reserve.name,
                    property,
                    value,
                    object_class=ClassEnum.Reserve,
                    collection=CollectionEnum.Reserves,
                    scenario=self.plexos_scenario,
                )
            time_series_properties = self._get_time_series_properties(reserve)
            if time_series_properties:
                text = time_series_properties.pop("Data File")
                for property_name, property_value in time_series_properties.items():
                    self._db_mgr.add_property(
                        reserve.name,
                        property_name,
                        property_value,
                        object_class=ClassEnum.Reserve,
                        collection=CollectionEnum.Reserves,
                        scenario=self.plexos_scenario,
                        text={"Data File": text},
                    )

            # Add Regions properties. Currently, we only add the load_risk
            component_dict = reserve.model_dump(
                exclude_none=True, exclude=NESTED_ATTRIBUTES | {"max_requirement"}
            )

            if not reserve.region:
                return
            reserve_region = reserve.region
            assert reserve_region is not None
            regions = self.system.get_components(
                ACBus, filter_func=lambda x: x.load_zone.name == reserve_region.name
            )

            collection_properties = self._db_mgr.get_valid_properties(
                collection=CollectionEnum.Regions,
                parent_class=ClassEnum.Reserve,
                child_class=ClassEnum.Region,
            )
            for region in regions:
                self._db_mgr.add_membership(
                    reserve.name,
                    region.name,  # Zone has the same name
                    parent_class=ClassEnum.Reserve,
                    child_class=ClassEnum.Region,
                    collection=CollectionEnum.Regions,
                )
                properties = get_export_properties(
                    component_dict,
                    partial(apply_property_map, property_map=self.property_map),
                    partial(apply_pint_deconstruction, unit_map=self.default_units),
                    partial(apply_valid_properties, valid_properties=collection_properties),
                )
                if properties:
                    for property_name, property_value in properties.items():
                        self._db_mgr.add_property(
                            region.name,
                            property_name,
                            property_value,
                            object_class=ClassEnum.Region,
                            parent_object_name=reserve.name,
                            parent_class=ClassEnum.Reserve,
                            collection=CollectionEnum.Regions,
                            scenario=self.plexos_scenario,
                        )
        return

    def add_generators(self):
        """Add generator objects to the database."""

        # Add generator objects excluding batteries
        def exclude_battery(component):
            return not isinstance(component, EnergyReservoirStorage)

        self.add_component_category(Generator, class_enum=ClassEnum.Generator, filter_func=exclude_battery)
        self.bulk_insert_objects(
            Generator,
            class_enum=ClassEnum.Generator,
            collection_enum=CollectionEnum.Generators,
            filter_func=exclude_battery,
        )
        self.insert_component_properties(
            Generator,
            parent_class=ClassEnum.System,
            collection=CollectionEnum.Generators,
            filter_func=exclude_battery,
            exclude_fields=[],
        )

        # Add generator memberships
        logger.debug("Adding generator memberships")
        for generator in self.system.get_components(Generator, filter_func=exclude_battery):
            self._db_mgr.add_membership(
                generator.name,
                generator.bus.name,
                parent_class=ClassEnum.Generator,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.Nodes,
            )
            properties = self._get_time_series_properties(generator)
            if properties:
                text = properties.pop("Data File")
                for property_name, property_value in properties.items():
                    self._db_mgr.add_property(
                        generator.name,
                        property_name,
                        property_value,
                        object_class=ClassEnum.Generator,
                        collection=CollectionEnum.Generators,
                        scenario=self.plexos_scenario,
                        text={"Data File": text},
                    )
            if generator.services:
                for service in generator.services:
                    match service:
                        case Reserve():
                            self._db_mgr.add_membership(
                                service.name,
                                generator.name,
                                parent_class=ClassEnum.Reserve,
                                child_class=ClassEnum.Generator,
                                collection=CollectionEnum.Generators,
                            )
                        case _:
                            raise NotImplementedError(f"{service} not yet implemented for generator.")

        # NOTE: This needs to be optimized. It is currently slow.
        logger.debug("Adding generator emisssions memberships")
        for emission in self.system.get_supplemental_attributes(Emission):
            for generator in self.system.get_components_with_supplemental_attribute(emission):
                self._db_mgr.add_membership(
                    emission.emission_type,
                    generator.name,
                    parent_class=ClassEnum.Emission,
                    child_class=ClassEnum.Generator,
                    collection=CollectionEnum.Generators,
                )
                self._db_mgr.add_property(
                    generator.name,
                    self.property_map["rate"],
                    get_magnitude(emission.rate),
                    object_class=ClassEnum.Generator,
                    parent_class=ClassEnum.Emission,
                    parent_object_name=emission.emission_type,
                    collection=CollectionEnum.Generators,
                    scenario=self.plexos_scenario,
                )

    def add_batteries(self):
        """Add battery objects to the database."""
        # Add battery objects
        self.add_component_category(EnergyReservoirStorage, class_enum=ClassEnum.Battery)
        self.bulk_insert_objects(
            EnergyReservoirStorage,
            class_enum=ClassEnum.Battery,
            collection_enum=CollectionEnum.Batteries,
        )
        self.insert_component_properties(
            EnergyReservoirStorage, parent_class=ClassEnum.System, collection=CollectionEnum.Batteries
        )
        # Add battery memberships
        for battery in self.system.get_components(EnergyReservoirStorage):
            self._db_mgr.add_membership(
                battery.name,
                battery.bus.name,
                parent_class=ClassEnum.Battery,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.Nodes,
            )
            if battery.services:
                for service in battery.services:
                    match service:
                        case Reserve():
                            self._db_mgr.add_membership(
                                service.name,
                                battery.name,
                                parent_class=ClassEnum.Reserve,
                                child_class=ClassEnum.Battery,
                                collection=CollectionEnum.Batteries,
                            )
                        case _:
                            raise NotImplementedError(f"{service} not yet implemented for generator.")

    def add_storage(self):
        """Add storage objects to the database."""
        # Add pump storage objects
        self.add_component_category(
            HydroPumpedStorage, category_attribute="head", class_enum=ClassEnum.Storage
        )
        self.add_component_category(
            HydroPumpedStorage, category_attribute="tail", class_enum=ClassEnum.Storage
        )
        self.bulk_insert_objects(
            HydroPumpedStorage,
            class_enum=ClassEnum.Storage,
            category_attribute="head",
            collection_enum=CollectionEnum.Storages,
            name_prefix="_head",
        )
        self.bulk_insert_objects(
            HydroPumpedStorage,
            class_enum=ClassEnum.Storage,
            category_attribute="tail",
            collection_enum=CollectionEnum.Storages,
            name_prefix="_tail",
        )

        head_storage = [
            component.model_dump(exclude_none=True, exclude=NESTED_ATTRIBUTES)
            for component in self.system.get_components(HydroPumpedStorage)
        ]
        for component in head_storage:
            component["name"] += "_head"
        self.insert_component_properties(
            HydroPumpedStorage,
            parent_class=ClassEnum.System,
            collection=CollectionEnum.Storages,
            records=head_storage,
        )
        tail_storage = [
            component.model_dump(exclude_none=True, exclude=NESTED_ATTRIBUTES)
            for component in self.system.get_components(HydroPumpedStorage)
        ]
        for component in tail_storage:
            component["name"] += "_tail"
        self.insert_component_properties(
            HydroPumpedStorage,
            parent_class=ClassEnum.System,
            collection=CollectionEnum.Storages,
            records=tail_storage,
        )

        for phs in self.system.get_components(HydroPumpedStorage):
            head_name = f"{phs.name}_head"
            tail_name = f"{phs.name}_tail"
            self._db_mgr.add_membership(
                phs.name,
                head_name,
                parent_class=ClassEnum.Generator,
                child_class=ClassEnum.Storage,
                collection=CollectionEnum.HeadStorage,
            )
            self._db_mgr.add_membership(
                phs.name,
                tail_name,
                parent_class=ClassEnum.Generator,
                child_class=ClassEnum.Storage,
                collection=CollectionEnum.TailStorage,
            )
        return

    def _add_simulation_objects(self):
        for simulation_object in self.simulation_objects:
            collection_enum = get_enum_from_string(simulation_object["collection_name"], CollectionEnum)
            class_enum: ClassEnum = get_enum_from_string(simulation_object["class_name"], ClassEnum)
            for objects in simulation_object["attributes"]:
                self._db_mgr.add_object(
                    objects["name"],
                    class_enum,
                    collection_enum,
                    category_name=objects["category"] or "-",
                )

            # Add attributes
            property_list = self.output_config.defaults[simulation_object["class_name"]]
            for attributes in property_list["attributes"]:
                self._db_mgr.add_attribute(
                    object_name=attributes["name"],
                    object_class=get_enum_from_string(attributes["class"], ClassEnum),
                    attribute_class=class_enum,
                    attribute_name=attributes["attribute"],
                    attribute_value=attributes["value"],
                )

    def _add_horizons(self):
        logger.info("Adding model horizon")
        for horizon, values in self.static_horizons.items():
            self._db_mgr.add_object(
                horizon,
                ClassEnum.Horizon,
                CollectionEnum.Horizons,
            )
            for attribute, attribute_value in values["attributes"].items():
                self._db_mgr.add_attribute(
                    object_name=horizon,
                    object_class=ClassEnum.Horizon,
                    attribute_class=ClassEnum.Horizon,
                    attribute_name=attribute,
                    attribute_value=attribute_value,
                )

    def _add_models(self):
        for model, values in self.static_models.items():
            self._db_mgr.add_object(
                model,
                ClassEnum.Model,
                CollectionEnum.Models,
                category_name=values["category"],
            )
            for attribute, attribute_value in values["attributes"].items():
                self._db_mgr.add_attribute(
                    object_name=model,
                    object_class=ClassEnum.Model,
                    attribute_class=ClassEnum.Model,
                    attribute_name=attribute,
                    attribute_value=attribute_value,
                )
            for child_class_name, child_object_name in values["memberships"].items():
                collection_enum = get_enum_from_string(child_class_name, CollectionEnum)
                child_class = get_enum_from_string(child_class_name, ClassEnum)
                self._db_mgr.add_membership(
                    model,
                    child_object_name,
                    parent_class=ClassEnum.Model,
                    child_class=child_class,
                    collection=collection_enum,
                )
            self._db_mgr.add_membership(
                model,
                self.plexos_scenario,
                parent_class=ClassEnum.Model,
                child_class=ClassEnum.Scenario,
                collection=CollectionEnum.Scenarios,
            )
        return

    def _add_reports(self):
        logger.debug("Using {} for reports.", self.plexos_reports_fpath)
        report_objects = read_json(self.plexos_reports_fpath)

        for report_object in report_objects:
            report_object["collection"] = get_enum_from_string(report_object["collection"], CollectionEnum)
            report_object["parent_class"] = get_enum_from_string(report_object["parent_class"], ClassEnum)
            report_object["child_class"] = get_enum_from_string(report_object["child_class"], ClassEnum)
            self._db_mgr.add_report(**report_object)
        return

    def _get_category_id(self, component, category_attribute, categories_ids, default_category="-"):
        category_to_get = custom_attrgetter(component, category_attribute)

        if category_to_get not in categories_ids:
            return categories_ids[default_category]
        return categories_ids[category_to_get]


def apply_operation_cost(component: dict) -> dict[str, Any]:
    """Parse Infrasys Operation Cost into Plexos Records."""
    if not (cost := component.get("operation_cost")):
        return component
    match cost["class_type"]:
        case "ThermalGenerationCost":
            if shut_down := cost.get("start_up"):
                component["Start Cost"] = shut_down
            if shut_down := cost.get("shut_down"):
                component["Shutdown Cost"] = shut_down

            if cost.get("variable"):
                component = _variable_type_parsing(component, cost)
        case _:
            pass
    return component


def _variable_type_parsing(component: dict, cost_dict: dict[str, Any]) -> dict[str, Any]:
    fuel_curve = cost_dict["variable"]
    value_curve_type = cost_dict["value_curve_type"]
    variable_type = cost_dict["variable_type"]
    function_data = fuel_curve["value_curve"]["function_data"]
    match variable_type:
        case "FuelCurve":
            match value_curve_type:
                case "AverageRateCurve":
                    component["Heat Rate"] = function_data.proportional_term
                case "InputOutputCurve":
                    raise NotImplementedError("`InputOutputCurve` not yet implemented on Plexos exporter.")
        case "CostCurve":
            pass

    if fuel_cost := fuel_curve.get("fuel_cost"):
        component["Fuel Price"] = fuel_cost
    if vom_cost := fuel_curve.get("vom_cost"):
        component["VO&M Charge"] = vom_cost["function_data"].proportional_term

    return component



"""Helper functions for base parser."""

import os
from pathlib import Path

import h5py
import pandas as pd
import polars as pl
from loguru import logger

from .polars_helpers import pl_lowercase


def csv_handler(fpath: Path, csv_file_encoding="utf8", **kwargs) -> pl.DataFrame:
    """Parse CSV files and return a Polars DataFrame with all column names in lowercase.

    Parameters
    ----------
    fpath : str
        The file path of the CSV file to read.
    csv_file_encoding : str, optional
        The encoding format of the CSV file, by default "utf8".
    **kwargs : dict, optional
        Additional keyword arguments passed to the `pl.read_csv` function.

    Returns
    -------
    pl.DataFrame or None
        The parsed CSV file as a Polars DataFrame with lowercase column names if successful,
        or `None` if the file was not found.

    Raises
    ------
    pl.exceptions.ComputeError
        Raised if there are issues with the data types in the CSV file.
    FileNotFoundError
        Raised if the file is not found.

    See Also
    --------
    pl_lowercase : Function to convert all column names of a Polars DataFrame to lowercase.

    Example
    -------
    >>> df = csv_handler("data/example.csv")
    >>> print(df)
    shape: (2, 3)
    ┌─────┬────────┬──────┐
    │ id  │ name   │ age  │
    │ --- │ ---    │ ---  │
    │ i64 │ str    │ i64  │
    ╞═════╪════════╪══════╡
    │ 1   │ Alice  │ 30   │
    │ 2   │ Bob    │ 24   │
    └─────┴────────┴──────┘
    """
    logger.trace("Attempting reading file {}", fpath)
    logger.trace("Parsing file {}", fpath)
    try:
        data_file = pl.read_csv(
            fpath.as_posix(),
            infer_schema_length=10_000_000,
            encoding=csv_file_encoding,
        )
    except FileNotFoundError:
        msg = f"File {fpath} not found."
        logger.error(msg)
        raise FileNotFoundError(msg)
    except pl.exceptions.PolarsError:
        logger.warning("File {} could not be parse due to dtype problems. See error.", fpath)
        raise

    if kwargs.get("keep_case") is None:
        data_file = pl_lowercase(data_file)

    return data_file


def h5_handler(fpath, parser_class: str, **kwargs) -> pl.LazyFrame:
    """Parse H5 files and return a Polars DataFrame.

    Currently, the only exception we handle is for ReEDS since it is formatting differently. If new parsers or
    existing parsers required h5 reading, we will need to add new handlers.

    Parameters
    ----------
    fpath : str
        The file path of the CSV file to read.
    csv_file_encoding : str, optional
        The encoding format of the CSV file, by default "utf8".

    Raises
    ------
    NotImplementedError
        Raised if a non supported parser request a h5 file.
    """
    match parser_class:
        case "ReEDSParser":
            with h5py.File(fpath, "r") as f:
                logger.info(f"Parsing h5 File: {fpath}")
                # unique structure for recf versus load
                if os.path.basename(fpath) == "recf.h5":
                    index_datetime = f["index_datetime"][:]
                    pd_df = pd.DataFrame(
                        f["data"],
                        index=pd.Index(
                            pd.to_datetime(
                                index_datetime.astype(str),
                            ),
                            name="datetime",
                        ),
                        columns=[col.decode("utf-8") for col in f["columns"]],
                    )
                    pd_df = pd_df.reset_index()
                elif os.path.basename(fpath) == "load.h5":
                    # get index data
                    index_year = f["index_year"][:]
                    index_datetime = f["index_datetime"][:]
                    # get data in pandas format
                    pd_df = pd.DataFrame(f["data"], columns=[col.decode("utf-8") for col in f["columns"]])

                    # creating multi level index
                    unique_year_sorted = sorted(set(index_year))
                    unique_datetime_sorted = sorted(set(index_datetime))
                    multilevel_index = pd.MultiIndex.from_product(
                        [unique_year_sorted, unique_datetime_sorted], names=["year", "datetime"]
                    )

                    # setting index then reseting
                    pd_df = pd_df.set_index(multilevel_index).reset_index()

                    # correctly formating datetime col
                    pd_df["datetime"] = pd.to_datetime(pd_df["datetime"].astype(str))
                pl_df = pl.LazyFrame(pd_df)
        case _:
            msg = f"H5 file parsing is not implemented for {parser_class=}."
            raise NotImplementedError(msg)

    if kwargs.get("keep_case") is None:
        pl_df = pl_lowercase(pl_df)
    return pl_df



from r2x.api import System


def test_serialization(infrasys_test_system, tmp_path):
    system = infrasys_test_system
    system.to_json(tmp_path / "test.json")
    deserialized_system = System.from_json(tmp_path / "test.json")

    assert system._uuid == deserialized_system._uuid
    assert system._components.get_num_components() == deserialized_system._components.get_num_components()



from r2x.enums import PrimeMoversType
from r2x.models import Generator, ACBus, Emission, HydroPumpedStorage, ThermalStandard
from r2x.models import MinMax
from r2x.parser.handler import create_model_instance
from r2x.units import EmissionRate, ureg


def test_generator_model():
    generator = ThermalStandard.example()
    assert isinstance(generator, ThermalStandard)
    assert isinstance(generator.active_power.magnitude, float)


def test_emission_objects():
    emission = Emission.example()

    assert isinstance(emission, Emission)
    assert isinstance(emission.rate, EmissionRate)
    assert isinstance(emission.emission_type, str)
    assert emission.emission_type == "CO2"


def test_bus_model():
    bus = ACBus.example()
    assert isinstance(bus, ACBus)
    assert isinstance(bus.number, int)


def test_generator_objects():
    bus = ACBus.example()
    generator = Generator(name="GEN01", active_power=100 * ureg.MW, bus=bus)
    assert isinstance(generator.bus, ACBus)


def test_pumped_hydro_generator():
    pumped_storage = HydroPumpedStorage.example()
    assert isinstance(pumped_storage, HydroPumpedStorage)
    assert isinstance(pumped_storage.bus, ACBus)
    assert isinstance(pumped_storage.prime_mover_type, PrimeMoversType)
    assert pumped_storage.prime_mover_type == PrimeMoversType.PS


def test_serialize_active_power_limits():
    active_power_limits = MinMax(min=0, max=100)
    generator = Generator(name="TestGEN", active_power_limits=active_power_limits)

    output = generator.model_dump()
    assert output["active_power_limits"] == {"min": 0, "max": 100}

    output = generator.serialize_active_power_limits(active_power_limits)
    assert output == {"min": 0, "max": 100}


def test_create_model_instance():
    name = "TestGen"
    generator = create_model_instance(Generator, name=name)
    assert isinstance(generator, Generator)
    assert isinstance(generator.name, str)
    assert generator.name == name

    name = ["TestGen"]
    generator = create_model_instance(Generator, name=name, skip_validation=True)
    assert isinstance(generator, Generator)
    assert isinstance(generator.name, list)
    assert generator.name == name
